import os, random, shlex, shutil, sys, time, webbrowser

from xklb.files import sample_compare
from xklb.utils import arggroups, consts, file_utils, path_utils, processes, strings
from xklb.utils.log_utils import log

webbrowser.register("termux-open-url '%s'", None)


def get_ip_of_chromecast(device_name) -> str:
    from pychromecast import discovery

    cast_infos, browser = discovery.discover_listed_chromecasts(friendly_names=[device_name])
    browser.stop_discovery()
    if not cast_infos:
        log.error("Target chromecast device not found")
        raise SystemExit(53)

    return cast_infos[0].host


def clear_input() -> None:
    if consts.IS_LINUX:
        from termios import TCIFLUSH, tcflush

        tcflush(sys.stdin, TCIFLUSH)
    elif consts.IS_MAC:
        import select

        while select.select([sys.stdin], [], [], 0.01) == ([sys.stdin], [], []):
            sys.stdin.read(1)
    elif consts.IS_WINDOWS:
        if getattr(clear_input, "kbhit", None) is None:
            from msvcrt import getch, kbhit  # type: ignore

            clear_input.kbhit = kbhit
            clear_input.getch = getch

        # Try to flush the buffer
        while clear_input.kbhit():
            clear_input.getch()


class InteractivePrompt(Exception):
    pass


def confirm(*args, **kwargs) -> bool:
    from rich.prompt import Confirm

    if consts.PYTEST_RUNNING:
        raise InteractivePrompt

    clear_input()
    return Confirm.ask(*args, **kwargs, default=False)


def preserve_root(p):
    assert p != os.sep
    drive, realpath = os.path.splitdrive(p)
    assert realpath
    assert os.path.normpath(p) != os.path.normpath(drive)


def rename(args, src, dst):
    if args.simulate:
        print("rename", src, dst)
    else:
        log.debug("rename\t%s\t%s", src, dst)
        os.rename(src, dst)


def unlink(args, p):
    if args.simulate:
        print("unlink", p)
    else:
        log.debug("unlink\t%s", p)
        os.unlink(p)


def rmtree(args, p):
    if args.simulate:
        print("rmtree", p)
    elif os.path.isdir(p):
        log.debug("rmtree\t%s", p)
        shutil.rmtree(p)
    else:
        log.debug("unlink\t%s", p)
        os.unlink(p)


def log_size_diff(src_size, dst_size):
    src_size_str = strings.file_size(src_size)
    dst_size_str = strings.file_size(dst_size)
    diff_size_str = strings.file_size(abs(src_size - dst_size))
    if src_size == dst_size:
        print(f"Source and destination are the same size: {src_size_str}")
    elif src_size > dst_size:
        print(f"Source ({src_size_str}) is {dst_size_str} larger than destination ({diff_size_str})")
    elif src_size < dst_size:
        print(f"Source ({src_size_str}) is {dst_size_str} smaller than destination ({diff_size_str})")


def clobber(args, source, destination) -> tuple[str | None, str]:
    if source == destination:
        log.info("Destination is the same as source\t%s", destination)
        return None, destination

    if getattr(args, "skip_open", False) and file_utils.is_file_open(source):
        log.info("Source already has an open file handler\t%s", destination)
        return None, destination

    orig_destination = destination

    if os.path.exists(destination):
        if os.path.isdir(destination):
            log.info("File Over Folder conflict\t%s\t%s", source, destination)
            match args.file_over_folder:
                case arggroups.FileOverFolder.SKIP:
                    source = None
                case arggroups.FileOverFolder.RENAME_SRC:
                    destination = file_utils.alt_name(destination)
                case arggroups.FileOverFolder.RENAME_DEST:
                    existing_rename = file_utils.alt_name(destination)
                    rename(args, destination, existing_rename)
                case arggroups.FileOverFolder.DELETE_SRC:
                    unlink(args, source)
                    source = None
                case arggroups.FileOverFolder.DELETE_DEST:
                    rmtree(args, destination)
                case arggroups.FileOverFolder.MERGE:
                    destination = os.path.join(destination, path_utils.basename(destination))  # down
                    log.info("re-targeted %s -> %s", orig_destination, destination)
                    return clobber(args, source, destination)

        else:
            src_stat = os.stat(source)
            dst_stat = os.stat(destination)
            if os.path.samestat(src_stat, dst_stat):
                log.info("Destination is the same as source\t%s", destination)
                return None, destination

            src_size = src_stat.st_size
            dst_size = dst_stat.st_size

            if dst_size == 0:
                log.debug("Overwriting empty file destination %s\t%s", source, destination)
                unlink(args, destination)
                return source, destination

            log.info("File Over File conflict\t%s\t%s", source, destination)
            for s in args.file_over_file:
                log.debug(s)
                match s:
                    case arggroups.FileOverFileOptional.DELETE_DEST_HASH:
                        if sample_compare.sample_cmp(source, destination, ignore_holes=True):
                            unlink(args, destination)
                            break
                    case arggroups.FileOverFileOptional.DELETE_DEST_SIZE:
                        if src_size == dst_size:
                            unlink(args, destination)
                            break
                    case arggroups.FileOverFileOptional.DELETE_DEST_LARGER:
                        if src_size < dst_size:
                            unlink(args, destination)
                            break
                    case arggroups.FileOverFileOptional.DELETE_DEST_SMALLER:
                        if src_size > dst_size:
                            unlink(args, destination)
                            break
                    case arggroups.FileOverFileOptional.DELETE_SRC_HASH:
                        if sample_compare.sample_cmp(source, destination, ignore_holes=True):
                            unlink(args, source)
                            source = None
                            break
                    case arggroups.FileOverFileOptional.DELETE_SRC_SIZE:
                        if src_size == dst_size:
                            unlink(args, source)
                            source = None
                            break
                    case arggroups.FileOverFileOptional.DELETE_SRC_LARGER:
                        if src_size > dst_size:
                            unlink(args, source)
                            source = None
                            break
                    case arggroups.FileOverFileOptional.DELETE_SRC_SMALLER:
                        if src_size < dst_size:
                            unlink(args, source)
                            source = None
                            break
                    case arggroups.FileOverFile.SKIP:
                        source = None
                    case arggroups.FileOverFile.DELETE_DEST:
                        unlink(args, destination)
                    case arggroups.FileOverFile.DELETE_DEST_ASK:
                        print(source)
                        print("  -->", destination)
                        log_size_diff(src_size, dst_size)
                        if confirm("Replace destination file?"):
                            unlink(args, destination)
                    case arggroups.FileOverFile.RENAME_SRC:
                        destination = file_utils.alt_name(destination)
                    case arggroups.FileOverFile.RENAME_DEST:
                        existing_rename = file_utils.alt_name(destination)
                        rename(args, destination, existing_rename)
                    case arggroups.FileOverFile.DELETE_SRC:
                        unlink(args, source)
                        source = None

    else:
        parent_dir = os.path.dirname(destination)
        if parent_dir == "":  # relative file
            return source, destination

        try:
            os.makedirs(parent_dir, exist_ok=True)
        except (
            FileExistsError,
            NotADirectoryError,  # a file exists _somewhere_ in the path hierarchy
            FileNotFoundError,  # Windows
        ):
            parent_file = parent_dir
            while not os.path.exists(parent_file):  # until we find the file conflict
                parent_file = os.path.dirname(parent_file)  # up
                if parent_file == "":
                    log.info("Ran out of path from %s", parent_dir)
                    raise

            preserve_root(parent_file)

            log.warning("Folder Over File conflict %s\t%s", source, parent_file)
            match args.folder_over_file:
                case arggroups.FolderOverFile.SKIP:
                    source = None
                case arggroups.FolderOverFile.DELETE_SRC:
                    rmtree(args, source)
                    source = None
                case arggroups.FolderOverFile.DELETE_DEST:
                    unlink(args, parent_file)
                case arggroups.FolderOverFile.RENAME_DEST:
                    existing_rename = file_utils.alt_name(parent_file)
                    rename(args, parent_file, existing_rename)
                case arggroups.FolderOverFile.MERGE:
                    temp_rename = file_utils.alt_name(parent_file)
                    rename(args, parent_file, temp_rename)
                    os.makedirs(parent_dir, exist_ok=True)  # there can't be more than one blocking file

                    while os.path.exists(parent_file):  # until we find an open file slot
                        parent_file = os.path.join(parent_file, os.path.basename(parent_file))  # down

                    if source == os.path.commonpath([source, destination]):
                        # file was a conflict with destination path but let the caller rename it
                        return temp_rename, parent_file
                    rename(args, temp_rename, parent_file)  # temporary rename to final dest
                    if destination == parent_file:
                        log.info("re-targeted %s -> %s", orig_destination, destination)
                        return clobber(args, source, destination)

            if source:
                os.makedirs(parent_dir, exist_ok=True)  # original destination parent
        else:
            log.debug("Nothing to clobber %s\t%s", source, destination)

    if destination != orig_destination:
        log.info("re-targeted %s -> %s", orig_destination, destination)

    return source, destination


def prompt(*args, **kwargs) -> str:
    from rich.prompt import Prompt

    clear_input()
    return Prompt.ask(*args, **kwargs)


def set_readline_completion(list_) -> None:
    try:
        import readline
    except ModuleNotFoundError:
        # "Windows not supported"
        return

    def create_completer(list_):
        def list_completer(_text, state):
            line = readline.get_line_buffer()

            if not line:
                min_depth = min([s.count(os.sep) for s in list_]) + 1  # type: ignore
                result_list = [c + " " for c in list_ if c.count(os.sep) <= min_depth]
                random.shuffle(result_list)
                return result_list[:25][state]
            else:
                match_list = [s for s in list_ if s.startswith(line)]
                min_depth = min([s.count(os.sep) for s in match_list]) + 1  # type: ignore
                result_list = [c + " " for c in match_list if c.count(os.sep) <= min_depth]
                random.shuffle(result_list)
                return result_list[:15][state]

        return list_completer

    readline.set_completer(create_completer(list_))
    readline.set_completer_delims("\t")
    readline.parse_and_bind("tab: complete")
    return


def get_mount_stats(src_mounts) -> list[dict[str, int | float]]:
    mount_space = []
    total_used = 1
    total_free = 1
    grand_total = 1
    for src_mount in src_mounts:
        total, used, free = shutil.disk_usage(src_mount)
        total_used += used
        total_free += free
        grand_total += total
        mount_space.append((src_mount, used, free, total))

    return [
        {"mount": mount, "used": used / total_used, "free": free / total_free, "total": total / grand_total}
        for mount, used, free, total in mount_space
    ]


def browse(browser, urls):
    for url in urls:
        if not browser or browser in ["echo", "print"]:
            print(url)
        elif browser in ["default"]:
            browsing_success = webbrowser.open(url, new=2, autoraise=False)
            if browsing_success is False:
                log.info("Problem opening %s", url)
        else:
            processes.cmd_detach(*shlex.split(browser), url)

        if len(urls) >= consts.MANY_LINKS:
            time.sleep(1.5)



import contextlib, functools, importlib, json, multiprocessing, os, shlex, signal, subprocess, sys, threading
from contextlib import suppress
from pathlib import Path
from shutil import which
from typing import NoReturn

from xklb.utils import consts, iterables, nums
from xklb.utils.log_utils import log


def exit_nicely(_signal, _frame) -> NoReturn:
    log.warning("\nExiting... (Ctrl+C)")
    raise SystemExit(130)


signal.signal(signal.SIGINT, exit_nicely)


def no_media_found() -> NoReturn:
    log.error("No media found")
    raise SystemExit(2)


def player_exit(completed_process) -> NoReturn:
    log.error("Player exited with code %s", completed_process.returncode)
    log.debug(shlex.join(completed_process.args))
    raise SystemExit(completed_process.returncode)


def exit_error(msg) -> NoReturn:
    log.error("ERROR: " + msg)
    raise SystemExit(1)


def timeout(time_str) -> None:
    max_seconds = nums.human_to_seconds(time_str) or 0

    def exit_timeout(_signal, _frame):
        print(f"\nReached timeout... ({max_seconds}s)")
        raise SystemExit(124)

    signal.signal(signal.SIGALRM, exit_timeout)
    signal.alarm(max_seconds)


sizeout_max = None
sizeout_total = 0


def sizeout(max_size: str, next_size: int) -> None:
    global sizeout_max
    global sizeout_total

    if sizeout_max is None:
        sizeout_max = nums.human_to_bytes(max_size)

    sizeout_total += next_size
    if sizeout_total > sizeout_max:
        print(f"\nReached sizeout... ({max_size})")
        raise SystemExit(124)


def with_timeout(seconds):  # noqa: ANN201
    def decorator(decorated):
        @functools.wraps(decorated)
        def inner(*args, **kwargs):
            pool = multiprocessing.Pool(1)
            async_result = pool.apply_async(decorated, args, kwargs)
            try:
                return async_result.get(seconds)
            finally:
                pool.close()

        return inner

    return decorator


def with_timeout_thread(seconds):  # noqa: ANN201
    def decorator(decorated):
        @functools.wraps(decorated)
        def inner(*args, **kwargs):
            event = threading.Event()  # task completed

            result = None

            def target():
                nonlocal result
                result = decorated(*args, **kwargs)
                event.set()

            thread = threading.Thread(target=target)
            thread.start()

            if event.wait(seconds):
                return result
            else:
                raise TimeoutError

        return inner

    return decorator


@contextlib.contextmanager
def timeout_thread(seconds):
    event = threading.Event()  # task completed
    result = None

    def target(func, *args, **kwargs):
        nonlocal result
        result = func(*args, **kwargs)
        event.set()

    def run(func, *args, **kwargs):
        thread = threading.Thread(target=target, args=(func, *args), kwargs=kwargs)
        thread.start()
        if event.wait(seconds):
            return result
        else:
            raise TimeoutError

    yield run


def os_bg_kwargs() -> dict:
    # prevent ctrl-c from affecting subprocesses first

    if hasattr(os, "setpgrp"):
        return {"start_new_session": True}
    else:
        # CREATE_NEW_PROCESS_GROUP = 0x00000200
        # DETACHED_PROCESS = 0x00000008
        # os_kwargs = dict(creationflags=DETACHED_PROCESS | CREATE_NEW_PROCESS_GROUP)
        return {}


def cmd(
    *command, strict=True, cwd=None, quiet=True, error_verbosity=1, ignore_regexps=None, **kwargs
) -> subprocess.CompletedProcess:
    def print_std(s, is_success):
        if ignore_regexps is not None:
            s = "\n".join(line for line in s.splitlines() if not any(r.match(line) for r in ignore_regexps))

        s = s.strip()
        if s:
            if quiet and is_success:
                log.debug(s)
            elif consts.PYTEST_RUNNING:
                log.warning(s)
            elif error_verbosity == 0:
                log.debug(s)
            elif error_verbosity == 1:
                log.info(s)
            else:
                log.warning(s)
        return s

    try:
        r = subprocess.run(
            command,
            capture_output=True,
            text=True,
            cwd=cwd,
            errors=sys.getfilesystemencodeerrors(),
            **os_bg_kwargs(),
            **kwargs,
        )
    except UnicodeDecodeError:
        print(repr(command))
        raise

    log.debug(r.args)
    print_std(r.stdout, r.returncode == 0)
    print_std(r.stderr, r.returncode == 0)
    if r.returncode != 0:
        if error_verbosity == 0:
            log.debug("[%s] exited %s", shlex.join(command), r.returncode)
        elif error_verbosity == 1:
            log.info("[%s] exited %s", shlex.join(command), r.returncode)
        else:
            log.warning("[%s] exited %s", shlex.join(command), r.returncode)

        if strict:
            raise subprocess.CalledProcessError(r.returncode, shlex.join(command), r.stdout, r.stderr)

    return r


def cmd_detach(*command, **kwargs) -> subprocess.CompletedProcess:
    # https://stackoverflow.com/questions/62521658/python-subprocess-detach-a-process
    # After lb closes, the detached process becomes daemonized (ie. not connected to the terminal so they won't show up in the shell command `jobs`)
    # If you shut down your computer often, you may want to open: `watch progress -wc ffmpeg` in another terminal so that you don't forget many things are in the background
    # If using with ffmpeg remember to include ffmpeg's flag `-nostdin` in the command when calling this function
    stdout = subprocess.DEVNULL
    stderr = subprocess.DEVNULL
    stdin = subprocess.DEVNULL

    command = iterables.conform(command)
    if command[0] in ["fish", "bash"]:
        command = command[0:2] + [shlex.join(command[2:])]
    proc = subprocess.Popen(
        command, stdin=stdin, stdout=stdout, stderr=stderr, close_fds=True, **os_bg_kwargs(), **kwargs
    )
    log.debug("pid %s cmd: %s", proc.pid, command)
    return subprocess.CompletedProcess(command, 0, "Detached command is async")


def cmd_interactive(*command, strict=True) -> subprocess.CompletedProcess:
    return_code = os.spawnvpe(os.P_WAIT, command[0], command, os.environ)

    if return_code != 0:
        msg = f"[{shlex.join(command)}] exited {return_code}"
        if strict:
            raise RuntimeError(msg)
        else:
            log.info(msg)

    return subprocess.CompletedProcess(command, return_code)


def Pclose(process) -> subprocess.CompletedProcess:  # noqa: N802
    try:
        stdout, stderr = process.communicate(input)
    except subprocess.TimeoutExpired as exc:
        log.debug("subprocess.TimeoutExpired")
        process.kill()
        if consts.IS_WINDOWS:
            exc.stdout, exc.stderr = process.communicate()
        else:
            process.wait()
        raise
    except Exception as e:
        log.debug(e)
        process.kill()
        raise
    return_code = process.poll()
    return subprocess.CompletedProcess(process.args, return_code, stdout, stderr)


def load_or_install_modules(module_list):
    for modules in module_list:
        loaded = False
        for module_name in modules:
            try:
                importlib.import_module(module_name)
            except ImportError:
                pass
            else:
                loaded = True
                break

        if not loaded:
            for module_name in modules:
                try:
                    subprocess.check_call([sys.executable, "-m", "pip", "install", module_name])
                except subprocess.CalledProcessError:
                    pass
                else:
                    try:
                        importlib.import_module(module_name)
                    except ImportError:
                        pass
                    else:
                        loaded = True
                        break

        if not loaded:
            print(f"None of the python packages {modules} could be imported or installed")


class UnplayableFile(RuntimeError):
    pass


def is_album_art(s):
    from yt_dlp.utils import traverse_obj

    return traverse_obj(s, ["disposition", "attached_pic"]) == 1


class FFProbe:
    def __init__(self, path, *args):
        args = [
            "ffprobe",
            "-hide_banner",
            "-rw_timeout",
            "100000000",
            "-timeout",
            "45000000",
            "-show_format",
            "-show_streams",
            "-show_chapters",
            "-of",
            "json",
            *args,
            path,
        ]
        p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        out, err = p.communicate()
        if p.returncode != 0:
            log.info("ffprobe %s out %s error %s", p.returncode, out, err)
            if p.returncode == -2:
                raise KeyboardInterrupt
            elif p.returncode == 127:  # Cannot open shared object file
                raise RuntimeError
            elif p.returncode == -6:  # Too many open files
                raise OSError
            else:
                raise UnplayableFile(out, err)
        d = json.loads(out.decode("utf-8"))

        self.path = path

        self.streams = d.get("streams")
        self.chapters = d.get("chapters")
        self.format = d.get("format")

        self.video_streams = []
        self.audio_streams = []
        self.subtitle_streams = []
        self.album_art_streams = []
        self.other_streams = []

        for s in self.streams:
            if "codec_type" not in s:
                continue
            elif s["codec_type"] == "video":
                if is_album_art(s):
                    self.album_art_streams.append(s)
                else:
                    self.video_streams.append(s)
            elif s["codec_type"] == "audio":
                self.audio_streams.append(s)
            elif s["codec_type"] == "subtitle":
                self.subtitle_streams.append(s)
            else:
                self.other_streams.append(s)

        self.has_video = len(self.video_streams) > 0
        self.has_audio = len(self.audio_streams) > 0

        self.duration = None
        with suppress(KeyError):
            self.duration = float(self.format["duration"])
        if self.duration is None:
            with suppress(IndexError, KeyError):
                self.duration = float(self.audio_streams[0]["duration"])
        if self.duration is None:
            with suppress(IndexError, KeyError):
                self.duration = float(self.video_streams[0]["duration"])

        if self.duration and self.duration > 0:
            start = nums.safe_float(self.format.get("start_time")) or 0
            if start > 0:
                end = nums.safe_float(self.format.get("end_time")) or 0
                if end > 0:
                    self.duration = start - end
                else:
                    self.duration -= start


def unar_out_path(archive_path):
    output_path = str(Path(archive_path).with_suffix(""))
    if output_path.endswith(tuple(f".{ext}" for ext in consts.ARCHIVE_EXTENSIONS)):
        output_path = str(Path(output_path).with_suffix(""))
    return output_path


def lsar(archive_path):
    if not which("lsar"):
        log.error("[%s]: The 'lsar' command is not available. Install 'unar' to check archives", archive_path)
        return []

    lsar_output = subprocess.run(["lsar", "-json", archive_path], stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    try:
        lsar_json = json.loads(lsar_output.stdout)
    except json.JSONDecodeError:
        log.warning("[%s]: Error parsing lsar output as JSON: %s", archive_path, lsar_output)
        return []

    if lsar_json.get("lsarError"):
        log.warning("[%s]: lsar error %s", archive_path, lsar_json["lsarError"])

    lsar_contents = lsar_json.get("lsarContents", [])
    if len(lsar_contents) == 0:
        log.info("[%s]: archive empty", archive_path)
        return []

    ar_size = os.stat(archive_path).st_size
    unar_out = unar_out_path(archive_path)
    archive_info_list = []
    for entry in lsar_contents:
        archive_info_list.append(
            {
                "archive_path": archive_path,
                "path": os.path.join(unar_out, entry.get("XADFileName")),
                "compressed_size": entry.get("XADCompressedSize") or ar_size // len(lsar_contents),
                "size": entry.get("XADFileSize"),
            }
        )

    return archive_info_list


def unar_delete(archive_path):
    output_path = unar_out_path(archive_path)
    if not os.path.exists(output_path):
        os.makedirs(output_path)

    original_stats = os.stat(archive_path)
    cmd("unar", "-quiet", "-force-rename", "-no-directory", "-output-directory", output_path, archive_path)
    os.utime(output_path, (original_stats.st_atime, original_stats.st_mtime))

    lsar_json = cmd("lsar", "-json", archive_path).stdout
    lsar_output = json.loads(lsar_json)
    part_files = lsar_output["lsarProperties"]["XADVolumes"]

    try:
        for part_file in part_files:
            part_file_path = os.path.join(os.path.dirname(archive_path), part_file)
            os.unlink(part_file_path)
    except Exception as e:
        print(f"Error deleting files: {e}")



class HTTPTooManyRequests(EnvironmentError):
    pass


class RecoverableError(RuntimeError):
    pass


class UnrecoverableError(RuntimeError):
    pass


def raise_for_status(status_code):
    if status_code == 429:
        raise HTTPTooManyRequests
    elif status_code in (404,):
        raise UnrecoverableError("HTTP404: HTTPNotFound")
    elif 400 <= status_code < 500:
        msg = f"HTTP{status_code}"
        raise RecoverableError(msg)
    elif 500 <= status_code < 600:
        msg = f"HTTP{status_code}"
        raise RecoverableError(msg)



import json

from xklb import usage
from xklb.utils import arg_utils, arggroups, argparse_utils, consts, devices, iterables, printing, strings, web
from xklb.utils.log_utils import log


def parse_args():
    parser = argparse_utils.ArgumentParser(usage=usage.extract_links)
    arggroups.extractor(parser)
    arggroups.requests(parser)
    arggroups.selenium(parser)
    arggroups.filter_links(parser)

    parser.add_argument("--download", action="store_true", help="Download filtered links")
    arggroups.download(parser)
    parser.set_defaults(profile=consts.DBType.filesystem)

    arggroups.debug(parser)
    arggroups.paths_or_stdin(parser)
    args = parser.parse_intermixed_args()
    arggroups.args_post(args, parser)

    arggroups.extractor_post(args)
    arggroups.filter_links_post(args)
    web.requests_session(args)  # prepare requests session
    arggroups.selenium_post(args)

    return args


def is_desired_url(args, link, link_text, before, after) -> bool:
    include_cond = all if args.strict_include else any
    exclude_cond = all if args.strict_exclude else any

    link_lower = link if args.case_sensitive else link.lower()
    link_text_lower = link_text if args.case_sensitive else link_text.lower()

    if args.path_include and not include_cond(inc in link_lower for inc in args.path_include):
        log.debug("no match path-include: %s", link_lower)
        return False
    if args.path_exclude and exclude_cond(ex in link_lower for ex in args.path_exclude):
        log.debug("matched path-exclude: %s", link_lower)
        return False

    if args.text_exclude and exclude_cond(ex in link_text_lower for ex in args.text_exclude):
        log.debug("matched text-exclude: %s", link_text_lower)
        return False
    if args.text_include and not include_cond(inc in link_text_lower for inc in args.text_include):
        log.debug("no match text-include: %s", link_text_lower)
        return False

    if args.before_exclude or args.before_include:
        if args.before_include and not before:
            return False

        before_text = before if args.case_sensitive else before.lower()

        if args.before_exclude and exclude_cond(ex in before_text for ex in args.before_exclude):
            log.debug("matched before-exclude: %s", before_text)
            return False
        if args.before_include and not include_cond(inc in before_text for inc in args.before_include):
            log.debug("no match before-include: %s", before_text)
            return False

        if args.before_exclude or args.before_include:  # just logging
            log.info("  before: %s", before_text)

    if args.after_exclude or args.after_include:
        if args.after_include and not after:
            return False

        after_text = after if args.case_sensitive else after.lower()

        if args.after_exclude and exclude_cond(ex in after_text for ex in args.after_exclude):
            log.debug("matched after-exclude: %s", after_text)
            return False
        if args.after_include and not include_cond(inc in after_text for inc in args.after_include):
            log.debug("no match after-include: %s", after_text)
            return False

        if args.after_exclude or args.after_include:  # just logging
            log.info("  after: %s", after_text)

    if args.text_exclude or args.text_include:  # just logging
        log.info("  text: `%s`", link_text_lower.strip())

    return True


def parse_inner_urls(args, url, markup):
    from bs4 import BeautifulSoup

    soup = BeautifulSoup(markup, "lxml")

    link_attrs = set()
    if args.href:
        link_attrs.add("href")
    if args.src:
        link_attrs.add("src")
    if args.url:
        link_attrs.add("url")
    if args.data_src:
        link_attrs.update({"data-src", "data-url", "data-original"})

    url_renames = args.url_renames.items()

    def delimit_fn(el):
        return any(el.has_attr(s) for s in link_attrs)

    tags = web.tags_with_text(soup, delimit_fn)
    for tag in tags:
        for attr_name, attr_value in tag.attrs.items():
            if attr_name not in link_attrs:
                continue

            attr_value = str(attr_value).strip()
            if attr_value and attr_value[0] != "#":
                link = web.construct_absolute_url(url, attr_value)
                link_text = strings.remove_consecutive_whitespace(tag.text.strip())

                if is_desired_url(args, link, link_text, tag.before_text, tag.after_text):
                    for k, v in url_renames:
                        link = link.replace(k, v)

                    yield {
                        "link": link,
                        "link_text": strings.strip_enclosing_quotes(link_text),
                        "before_text": tag.before_text,
                        "after_text": tag.after_text,
                    }


def get_inner_urls(args, url):
    log.debug("Loading links from %s", url)

    is_error = False
    if args.selenium:
        web.selenium_get_page(args, url)

        if args.manual:
            while devices.confirm("Extract HTML from browser?"):
                markup = web.selenium_extract_html(args.driver)
                yield from parse_inner_urls(args, url, markup)
        else:
            for markup in web.infinite_scroll(args.driver):
                yield from parse_inner_urls(args, url, markup)
    else:
        if args.local_html:
            with open(url) as f:
                markup = f.read()
            url = "file://" + url
        else:
            try:
                r = web.session.get(url, timeout=120)
            except Exception as e:
                if "too many 429 error" in str(e):
                    raise
                log.exception("Could not get a valid response from the server")
                return None
            if r.status_code == 404:
                log.warning("404 Not Found Error: %s", url)
                is_error = True
            else:
                r.raise_for_status()
            markup = r.content

        yield from parse_inner_urls(args, url, markup)

    web.sleep(args)

    if is_error:
        return None


def print_or_download(args, d):
    url = d["link"]
    if args.download:
        try:
            web.download_url(args, url)
        except RuntimeError as e:
            log.error("[%s]: %s", url, e)
    else:
        if not args.no_url_decode:
            url = web.url_decode(url).strip()
        if args.verbose >= consts.LOG_DEBUG:
            printing.pipe_print(json.dumps(d, ensure_ascii=False))
        else:
            printing.pipe_print(url)


def extract_links() -> None:
    args = parse_args()

    if args.insert_only:
        for url in arg_utils.gen_paths(args):
            if args.url_encode:
                url = web.url_encode(url).strip()
            if args.download:
                try:
                    web.download_url(args, url)
                except RuntimeError as e:
                    log.error("[%s]: %s", url, e)
            else:
                printing.pipe_print(url)
        return

    if args.selenium:
        web.load_selenium(args)
    try:
        for url in arg_utils.gen_paths(args):
            for d in iterables.return_unique(get_inner_urls, lambda d: d["link"])(args, url):
                print_or_download(args, d)

    finally:
        if args.selenium:
            web.quit_selenium(args)


if __name__ == "__main__":
    extract_links()



import json, random, time
from urllib.parse import parse_qs, urlencode, urlparse, urlunparse

import requests

from xklb import usage
from xklb.mediadb import db_media, db_playlists
from xklb.text import extract_links
from xklb.utils import arg_utils, arggroups, argparse_utils, consts, db_utils, objects, printing, strings, web
from xklb.utils.log_utils import log


def parse_args(action, **kwargs):
    parser = argparse_utils.ArgumentParser(**kwargs)
    parser.add_argument("--category", "-c")

    paging_parser = parser.add_argument_group("Paging")
    paging_parser.add_argument("--max-pages", type=int, help="Set a max number of pages to get")
    paging_parser.add_argument(
        "--fixed-pages",
        type=int,
        help="""Force a specific number of pages

Overrides --max-pages and --stop-known but you can still stop early via --stop-link ie. 429 page

If `--fixed-pages` is 1 and --start-page is not set then the URL will not be modified.

library links-add --fixed-pages=1
Loading page https://site/path

library links-add --fixed-pages=1 --page-start 99
Loading page https://site/path?page=99""",
    )
    paging_parser.add_argument(
        "--backfill-pages",
        "--backfill",
        type=int,
        help="""Similar to --fixed-pages but only for the first run

- Set `--backfill-pages` to the desired number of pages for the first run
- Set `--fixed-pages` to _always_ fetch the desired number of pages (remembered when using linksupdate)

If the website is supported by --auto-pager data is fetched twice when using page iteration.
As such, manual page iteration (--max-pages, --fixed-pages, etc) is disabled when using `--auto-pager`.

You can unset --fixed-pages for all the playlists in your database by running this command:
sqlite your.db "UPDATE playlists SET extractor_config = json_replace(extractor_config, '$.fixed_pages', null)"
""",
    )
    paging_parser.add_argument(
        "--page-step",
        "--step",
        type=int,
        default=1,
        help="""Use -1 for reverse paging

Some pages don't count page numbers but instead count items like messages or forum posts. You can iterate through like this:

library links-add --page-key start --page-start 0 --page-step 50

which translates to
&start=0    first page
&start=50   second page
&start=100  third page""",
    )
    paging_parser.add_argument("--page-start", "--start-page", "--start", type=int, help="Page number to start from")

    paging_parser.add_argument(
        "--page-key",
        default="page",
        help="""By default the script will attempt to modify each given URL with a specific query parameter, "&page=1".
Override like so:
library links-add --page-key p  # "&p=1"

Some websites use paths instead of query parameters. In this case the URL provided must include the matching page folder:
library links-add --page-key page https://website/page/1/
library links-add --page-key article https://website/article/1/
""",
    )
    paging_parser.add_argument(
        "--page-replace",
        help="""If you have more complicated needs you can replace the page number with a named variable:
library links-add --page-replace NUMBER https://site/with/complex-NUMBER
library links-add --page-replace NUMBER https://website/page/2?page=NUMBER
library links-add --page-replace NUMBER https://website/page/NUMBER?page=2
""",
    )

    paging_parser.add_argument(
        "--stop-pages-no-match",
        "--stop-no-match",
        type=int,
        default=4,
        help="""Some websites don't give an error when you try to access pages which don't exist.
To compensate for this the script will only continue fetching pages until there are neither new nor known links for four pages.""",
    )
    paging_parser.add_argument(
        "--stop-pages-no-new",
        "--stop-no-new",
        type=int,
        default=10,
        help="""After encountering ten pages with no new links we stop""",
    )
    paging_parser.add_argument(
        "--stop-new", type=int, help="Stop fetching pages when encountering fewer than or equal to N new links"
    )
    paging_parser.add_argument(
        "--stop-known", type=int, help="Stop fetching pages when encountering more than N known links"
    )
    paging_parser.add_argument("--stop-link", help="Stop fetching pages when hitting a specific link")

    arggroups.filter_links(parser)

    arggroups.requests(parser)
    arggroups.selenium(parser)
    arggroups.extractor(parser)

    parser.add_argument("--force", action="store_true")

    arggroups.debug(parser)

    arggroups.database(parser)
    if "add" in action:
        arggroups.paths_or_stdin(parser)
    args = parser.parse_intermixed_args()
    arggroups.args_post(args, parser, create_db=True)

    if args.auto_pager:
        args.fixed_pages = 1

    arggroups.extractor_post(args)
    arggroups.filter_links_post(args)
    web.requests_session(args)  # prepare requests session
    arggroups.selenium_post(args)

    return args


def add_playlist(args, path):
    info = {
        "hostname": urlparse(path).hostname,
        "category": getattr(args, "category", None) or "Uncategorized",
        "time_created": consts.APPLICATION_START,
        "extractor_config": args.extractor_config,
        "time_modified": 0,
        "time_deleted": 0,
    }
    return db_playlists.add(args, str(path), info, extractor_key="LinksDB")


def consolidate_media(args, path: str) -> dict:
    return {
        "playlists_id": getattr(args, "playlists_id", None),
        "path": path,
        "category": getattr(args, "category", None) or "Uncategorized",
        "time_created": consts.APPLICATION_START,
        "time_modified": 0,
        "time_deleted": 0,
    }


def add_media(args, variadic):
    for path_or_dict in variadic:
        if isinstance(path_or_dict, str):
            path = strings.strip_enclosing_quotes(path_or_dict)
            d = objects.dict_filter_bool(consolidate_media(args, path))
        else:
            d = consolidate_media(args, strings.strip_enclosing_quotes(path_or_dict.pop("path")))
            d = objects.dict_filter_bool({**d, **path_or_dict})

        db_media.add(args, d)


def set_page(input_string, page_key, page_number):
    parsed_url = urlparse(input_string)

    path_parts = parsed_url.path.split("/")

    if page_key in path_parts:
        page_key_index = path_parts.index(page_key)
        path_parts[page_key_index + 1] = str(page_number)  # Replace the page number at the next index
        updated_path = "/".join(path_parts)
        modified_url = urlunparse(
            (
                parsed_url.scheme,
                parsed_url.netloc,
                updated_path,
                parsed_url.params,
                parsed_url.query,
                parsed_url.fragment,
            )
        )
    else:
        query_params = parse_qs(parsed_url.query)
        query_params[page_key] = [str(page_number)]
        updated_query = urlencode(query_params, doseq=True)
        modified_url = urlunparse(
            (
                parsed_url.scheme,
                parsed_url.netloc,
                parsed_url.path,
                parsed_url.params,
                updated_query,
                parsed_url.fragment,
            )
        )

    return modified_url


def count_pages(args, page_limit):
    if args.page_start is None:
        page_start = 1 if args.page_step == 1 else 0
    else:
        page_start = args.page_start

    if page_limit:
        yield from range(page_start, page_start + (page_limit * args.page_step), args.page_step)
    else:
        page_num = page_start
        while True:
            yield page_num
            page_num += args.page_step


def update_category(args, path):
    with args.db.conn:
        args.db.conn.execute("UPDATE media SET category = ? WHERE path = ?", [args.category, path])


def extractor(args, playlist_path):
    known_media = set()
    new_media = set()
    end_of_playlist = False
    page_limit = args.backfill_pages or args.fixed_pages or args.max_pages

    page_count = 0
    page_count_since_match = 0
    page_count_since_new = 0
    for page_value in count_pages(args, page_limit):
        if end_of_playlist:
            break

        page_count += 1
        if page_count > 3:
            time.sleep(random.uniform(0.3, 4.55))

        if (page_limit == 1 and args.page_start is None) or (page_value == (args.page_start or 0) == 0):
            page_path = playlist_path
        elif args.page_replace:
            page_path = playlist_path.replace(args.page_replace, str(page_value))
        else:
            page_path = set_page(playlist_path, args.page_key, page_value)

        log.info("Loading page %s", page_path)
        page_known = set()
        page_new = {}
        try:
            for link_dict in extract_links.get_inner_urls(args, page_path):
                link = link_dict.pop("link")

                if link == args.stop_link:
                    end_of_playlist = True
                    break

                if link in page_known:
                    pass
                elif db_media.exists(args, link):
                    page_known.add(link)
                    if args.category:
                        update_category(args, link)
                else:
                    page_new[link] = objects.merge_dict_values_str(page_new.get(link) or {}, link_dict)

                printing.print_overwrite(f"Page {page_count} link scan: {len(page_new)} new [{len(page_known)} known]")

                if not (args.backfill_pages or args.fixed_pages):
                    if (args.stop_known and len(page_known) > args.stop_known) or (
                        args.stop_new and args.stop_new >= len(page_new)
                    ):
                        end_of_playlist = True
                        break
        except requests.HTTPError as e:
            log.error(e)

        add_media(args, [{"path": k, **v} for k, v in page_new.items()])

        new_media |= set(page_new.keys())
        known_media |= page_known

        if len(page_new) > 0:
            page_count_since_new = 0
        else:
            page_count_since_new += 1
        if len(page_new) > 0 or len(page_known) > 0:
            page_count_since_match = 0
        else:
            page_count_since_match += 1

        if not (args.backfill_pages or args.fixed_pages) and (
            page_count_since_new >= args.stop_pages_no_new or page_count_since_match >= args.stop_pages_no_match
        ):
            end_of_playlist = True
    print()

    print(
        f"{len(new_media)} new [{len(known_media)} known] in {page_count} pages (avg {len(new_media) // page_count} new [{len(known_media) // page_count} known])"
    )
    return len(new_media)


def links_add() -> None:
    args = parse_args(consts.SC.links_add, usage=usage.links_add)

    if args.insert_only:
        media_new = set()
        media_known = set()
        for p in arg_utils.gen_paths(args):
            if db_media.exists(args, p):
                media_known.add(p)
                if args.category:
                    update_category(args, p)
            else:
                add_media(args, [p])
                media_new.add(p)
            printing.print_overwrite(f"Link import: {len(media_new)} new [{len(media_known)} known]")
    else:
        if args.selenium:
            web.load_selenium(args)
        try:
            playlist_count = 0
            for playlist_path in arg_utils.gen_paths(args):
                args.playlists_id = add_playlist(args, playlist_path)
                extractor(args, playlist_path)

                if playlist_count > 3:
                    time.sleep(random.uniform(0.05, 2))
                playlist_count += 1

        finally:
            if args.selenium:
                web.quit_selenium(args)

    if not args.db["media"].detect_fts():
        db_utils.optimize(args)


def links_update() -> None:
    args = parse_args(consts.SC.links_update, usage=usage.links_update)

    link_playlists = db_playlists.get_all(
        args,
        order_by="""length(path)-length(REPLACE(path, '/', '')) desc
        , random()
        """,
    )

    selenium_needed = args.selenium or any(
        json.loads(d.get("extractor_config") or "{}").get("selenium") for d in link_playlists
    )
    if selenium_needed:
        web.load_selenium(args)

    try:
        playlist_count = 0
        for playlist in link_playlists:
            extractor_config = json.loads(playlist.get("extractor_config") or "{}")
            args_env = arg_utils.override_config(args, extractor_config)

            new_media = extractor(args_env, playlist["path"])

            if new_media > 0:
                db_playlists.decrease_update_delay(args, playlist["path"])
            else:
                db_playlists.increase_update_delay(args, playlist["path"])

            if playlist_count > 3:
                time.sleep(random.uniform(0.05, 2))
            playlist_count += 1

    finally:
        if selenium_needed:
            web.quit_selenium(args)


if __name__ == "__main__":
    links_add()



import errno, mimetypes, os, shlex, shutil, tempfile, time
from collections import Counter, namedtuple
from fnmatch import fnmatch
from functools import wraps
from io import StringIO
from pathlib import Path
from shutil import which

import urllib3

from xklb.utils import consts, file_utils, printing, processes, web
from xklb.utils.log_utils import log


def scan_stats(files: int, filtered_files: int, folders: int, filtered_folders: int):
    return (
        f"""Files: {files}{f' [{filtered_files} ignored]' if filtered_files else ''}"""
        f""" Folders: {folders}{f' [{filtered_folders} ignored]' if filtered_folders else ''}"""
    )


def rglob(
    base_dir: str | Path,
    extensions=None,  # None | Iterable[str]
    exclude=None,  # None | Iterable[str]
    include=None,  # None | Iterable[str]
) -> tuple[set[str], set[str], set[str]]:
    if extensions:
        extensions = tuple(s if s.startswith(".") else f".{s}" for s in extensions)

    files = set()
    filtered_files = set()
    filtered_folders = set()
    folders = set()
    stack = [base_dir]
    while stack:
        current_dir = stack.pop()
        try:
            scanned_dir = os.scandir(current_dir)
        except (FileNotFoundError, PermissionError):
            pass
        except OSError as e:
            if e.errno == 23:  # Too many open files
                raise e
            elif e.errno == 5:  # Input/output error
                log.exception("Input/output error: check dmesg. Skipping folder %s", current_dir)
            raise
        else:
            for entry in scanned_dir:
                if entry.is_dir(follow_symlinks=False):
                    if exclude and any(entry.name == pattern or fnmatch(entry.path, pattern) for pattern in exclude):
                        filtered_folders.add(entry.path)
                        continue
                    if include and not any(
                        entry.name == pattern or fnmatch(entry.path, pattern) for pattern in include
                    ):
                        filtered_folders.add(entry.path)
                        continue
                    folders.add(entry.path)
                    stack.append(entry.path)
                elif entry.is_symlink():
                    continue
                else:  # file or close enough
                    if extensions and not entry.path.lower().endswith(extensions):
                        filtered_files.add(entry.path)
                        continue
                    if include and not any(
                        entry.name == pattern or fnmatch(entry.path, pattern) for pattern in include
                    ):
                        filtered_files.add(entry.path)
                        continue
                    if exclude and any(entry.name == pattern or fnmatch(entry.path, pattern) for pattern in exclude):
                        filtered_files.add(entry.path)
                        continue
                    files.add(entry.path)

            printing.print_overwrite(
                f"[{base_dir}] {scan_stats(len(files), len(filtered_files), len(folders), len(filtered_folders))}"
            )

    if not consts.PYTEST_RUNNING:
        print(f"\r[{base_dir}] {scan_stats(len(files), len(filtered_files), len(folders), len(filtered_folders))}")

    filtered_extensions = Counter(Path(s).suffix for s in filtered_files)
    log.info("Filtered extensions: %s", filtered_extensions)

    return files, filtered_files, folders


def rglob_gen(
    base_dir: str | Path,
    extensions=None,  # None | Iterable[str]
    exclude=None,  # None | Iterable[str]
    include=None,  # None | Iterable[str]
):
    if extensions:
        extensions = tuple(s if s.startswith(".") else f".{s}" for s in extensions)

    folders = set()
    stack = [base_dir]
    while stack:
        current_dir = stack.pop()
        try:
            scanned_dir = os.scandir(current_dir)
        except (FileNotFoundError, PermissionError):
            pass
        except OSError as e:
            if e.errno == 23:  # Too many open files
                raise e
            elif e.errno == 5:  # Input/output error
                log.exception("Input/output error: check dmesg. Skipping folder %s", current_dir)
            raise
        else:
            for entry in scanned_dir:
                if entry.is_dir(follow_symlinks=False):
                    if exclude and any(entry.name == pattern or fnmatch(entry.path, pattern) for pattern in exclude):
                        continue
                    if include and not any(
                        entry.name == pattern or fnmatch(entry.path, pattern) for pattern in include
                    ):
                        continue
                    folders.add(entry.path)
                    stack.append(entry.path)
                elif entry.is_symlink():
                    continue
                else:  # file or close enough
                    if extensions and not entry.path.lower().endswith(extensions):
                        continue
                    if include and not any(
                        entry.name == pattern or fnmatch(entry.path, pattern) for pattern in include
                    ):
                        continue
                    if exclude and any(entry.name == pattern or fnmatch(entry.path, pattern) for pattern in exclude):
                        continue
                    yield entry.path


def file_temp_copy(src) -> str:
    fo_dest = tempfile.NamedTemporaryFile(delete=False)
    with open(src, "r+b") as fo_src:
        shutil.copyfileobj(fo_src, fo_dest.file)
    fo_dest.seek(0)
    fname = fo_dest.name
    fo_dest.close()
    return fname


def trash(args, path: Path | str, detach=True) -> None:
    if Path(path).exists():
        trash_put = which(args.override_trash)
        if trash_put is not None:
            if not detach:
                processes.cmd(trash_put, path, strict=False)
                return
            try:
                processes.cmd_detach(trash_put, path)
            except Exception:
                processes.cmd(trash_put, path, strict=False)
        else:
            Path(path).unlink(missing_ok=True)


def is_file_open(path):
    if os.name == "nt":
        try:
            os.open(path, os.O_RDWR | os.O_EXCL)
            return False
        except OSError:
            return True
    else:
        open_files = set()
        for proc in os.listdir("/proc"):
            try:
                fd_dir = os.path.join("/proc", proc, "fd")
                for fd in os.listdir(fd_dir):
                    link = os.readlink(os.path.join(fd_dir, fd))
                    if link.startswith("/") and link == path:
                        open_files.add(link)
            except OSError:
                continue
        return path in open_files


def filter_file(path, sieve) -> None:
    with open(path) as fr:
        lines = fr.readlines()
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as temp:
            temp.writelines(line for line in lines if line.rstrip() not in sieve)
            temp.flush()
            os.fsync(temp.fileno())
    shutil.copy(temp.name, path)
    Path(temp.name).unlink()


def tempdir_unlink(pattern):
    temp_dir = tempfile.gettempdir()
    cutoff = time.time() - 15 * 60  # 15 minutes in seconds
    for p in Path(temp_dir).glob(pattern):
        try:
            if p.stat().st_mtime < cutoff:
                p.unlink(missing_ok=True)
        except FileNotFoundError:  # glob->stat() racing
            pass


def resolve_absolute_path(s):
    p = Path(s).expanduser()
    if p.is_absolute():
        p = p.resolve()
        if p.exists():
            return str(p)
    return s  # relative path


def resolve_absolute_paths(paths):
    if paths is None:
        return paths
    return [resolve_absolute_path(s) for s in paths]


def fast_glob(path_dir, limit=100):
    files = []
    with os.scandir(path_dir) as entries:
        for entry in entries:
            if entry.is_file():
                files.append(entry.path)
                if len(files) == limit:
                    break
    return sorted(files)


def copy_file(source_file, destination_file, simulate=False):
    if simulate:
        print("cp", source_file, destination_file)
    else:
        try:
            shutil.copy2(source_file, destination_file)
        except OSError as e:
            if e.errno in (errno.ENOENT, errno.EXDEV):
                os.makedirs(os.path.dirname(destination_file), exist_ok=True)
                shutil.copy2(source_file, destination_file)  # try again
            else:
                raise


def rename_no_replace(src, dst):
    if os.path.exists(dst) and not os.path.isdir(dst):
        msg = f"The destination file {dst} already exists."
        raise FileExistsError(msg)
    os.rename(src, dst)


def rename_move_file(source_file, destination_file, simulate=False):
    if simulate:
        print("mv", source_file, destination_file)
    else:
        try:
            os.rename(source_file, destination_file)  # performance
        except OSError as e:
            if e.errno == errno.ENOENT:
                try:
                    os.makedirs(os.path.dirname(destination_file), exist_ok=True)
                    os.rename(source_file, destination_file)  # try again
                except OSError as e:
                    if e.errno == errno.EXDEV:  # Cross-device
                        shutil.move(source_file, destination_file)  # Fallback to shutil.move
                    else:
                        raise
            elif e.errno == errno.EXDEV:  # Cross-device
                shutil.move(source_file, destination_file)  # Fallback to shutil.move
            else:
                raise


def move_files(file_list):
    for existing_path, new_path in file_list:
        try:
            os.rename(existing_path, new_path)
        except Exception:
            try:
                parent_dir = os.path.dirname(new_path)
                os.makedirs(parent_dir, exist_ok=True)

                shutil.move(existing_path, new_path)
            except Exception:
                log.exception("Could not move %s", existing_path)


def move_files_bash(file_list):
    move_sh = """#!/bin/sh
existing_path=$1
new_path=$2

# Attempt to rename the file/directory
mv -Tn "$existing_path" "$new_path" 2>/dev/null

if [ $? -ne 0 ]; then
    mkdir -p $(dirname "$new_path")
    mv -Tn "$existing_path" "$new_path"
fi
"""
    move_sh_path = Path(tempfile.mktemp(dir=consts.TEMP_SCRIPT_DIR, prefix="move_", suffix=".sh"))
    move_sh_path.write_text(move_sh)
    move_sh_path.chmod(move_sh_path.stat().st_mode | 0o100)

    with tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".tsv") as temp:
        temp.writelines(
            f"{shlex.quote(existing_path)}\t{shlex.quote(new_path)}\n" for existing_path, new_path in file_list
        )
        temp.flush()
        os.fsync(temp.fileno())

        print(f"""### Move {len(file_list)} files to new folders: ###""")
        print(rf"PARALLEL_SHELL=sh parallel --colsep '\t' -a {temp.name} -j 20 {move_sh_path}")


def get_file_encoding(path):
    import chardet

    if path.startswith("http"):
        detector = chardet.UniversalDetector()
        response = web.session.get(path, stream=True)
        response.raw.decode_content = True

        num_bytes = 0
        for chunk in response.iter_content(chunk_size=16_384):
            if num_bytes > 1048576:  # 1MiB
                break
            detector.feed(chunk)
            if detector.done:
                break
            num_bytes += len(chunk)
        detector.close()

        encoding = detector.result["encoding"]
    else:
        with open(path, "rb") as f:
            sample = f.read(1048576)  # 1 MiB

        encoding = chardet.detect(sample)["encoding"]

    if encoding:
        log.info(f"The encoding of {path} is likely: {encoding}")
    return encoding


def head_stream(url, head_len):
    head_response = web.session.get(url, stream=True, timeout=1)
    head_response.raw.decode_content = True
    head_response.raise_for_status()

    return head_response.raw.read(head_len)


def foot_stream(url, foot_len):
    foot_response = web.session.get(url, stream=True, headers={"Range": f"bytes=-{foot_len}"}, timeout=1)
    foot_response.raw.decode_content = True

    return foot_response.raw.read(foot_len)


def head_foot_stream(url, head_len, foot_len):
    import io

    try:
        head_bytes = head_stream(url, head_len)
    except TimeoutError:
        head_bytes = b""

    try:
        foot_bytes = foot_stream(url, foot_len)
    except (TimeoutError, urllib3.exceptions.DecodeError):
        foot_bytes = b""

    stream = io.BytesIO(head_bytes + foot_bytes)
    return stream


@processes.with_timeout_thread(max(consts.REQUESTS_TIMEOUT) + 5)
def mimetype(path):
    import puremagic

    p = Path(path)

    file_type = None
    ext = puremagic.ext_from_filename(path)
    if ext in (".zarr", ".zarr/"):
        file_type = "Zarr"
    elif p.is_dir():
        file_type = "directory"
    else:
        file_type, encoding = mimetypes.guess_type(path, strict=False)

    if ext and file_type is None:
        pandas_ext = {
            ".dta": "Stata",
            ".xlsx": "Excel",
            ".xls": "Excel",
            ".json": "JSON",
            ".jsonl": "JSON Lines",
            ".ndjson": "JSON Lines",
            ".geojson": "GeoJSON",
            ".geojsonl": "GeoJSON Lines",
            ".ndgeojson": "GeoJSON Lines",
            ".hdf": "HDF5",
            ".feather": "Feather",
            ".parquet": "Parquet",
            ".sas7bdat": "SAS",
            ".sav": "SPSS",
            ".pkl": "Pickle",
            ".orc": "ORC",
        }
        file_type = pandas_ext.get(ext)

    if file_type is None:
        try:
            if path.startswith("http"):
                max_head = max([len(x.byte_match) + x.offset for x in puremagic.magic_header_array])
                max_foot = max([len(x.byte_match) + abs(x.offset) for x in puremagic.magic_footer_array])
                info = puremagic.magic_stream(head_foot_stream(path, max_head, max_foot), path)
            else:
                info = puremagic.magic_file(path)
            log.debug(info)
            file_type = info[0].name
        except (puremagic.PureError, IndexError, ValueError):
            if p.is_socket():
                file_type = "socket"
            elif p.is_fifo():
                file_type = "fifo"
            elif p.is_symlink():
                file_type = "symlink"
            elif p.is_block_device():
                file_type = "block device"
            elif p.is_char_device():
                file_type = "char device"
            try:
                if Path(path).stat().st_size == 0:
                    file_type = "empty file"
            except Exception:
                return None

        except FileNotFoundError:
            return None

    return file_type


def retry_with_different_encodings(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except UnicodeDecodeError as exc:
            original_exc = exc

        likely_encodings = []
        detected_encoding = get_file_encoding(args[0])
        if detected_encoding:
            likely_encodings.append(detected_encoding)

        for encoding in likely_encodings + consts.COMMON_ENCODINGS:
            try:
                kwargs["encoding"] = encoding
                return func(*args, **kwargs)
            except UnicodeDecodeError:
                pass

        raise original_exc  # If no encoding worked, raise the original exception

    return wrapper


NDF = namedtuple("NamedDataFrame", ["df_name", "df"])


@retry_with_different_encodings
def read_file_to_dataframes(
    path,
    table_name=None,
    table_index=None,
    start_row=None,
    end_row=None,
    order_by=None,
    encoding=None,
    mimetype=None,
    join_tables=False,
    transpose=False,
    skip_headers=False,
) -> list[NDF]:
    import pandas as pd

    if mimetype is None:
        mimetype = file_utils.mimetype(path)
    if mimetype is not None:
        mimetype = mimetype.strip().lower()
    log.info(mimetype)

    if mimetype is None:
        msg = f"{path}: File type could not be determined. Pass in --filetype"
        raise ValueError(msg)

    dfs: list[NDF] = []

    if mimetype in ("sqlite", "sqlite3", "sqlite database file"):
        import pandas as pd
        from sqlite_utils import Database

        db = Database(path)

        if table_name:
            tables = [table_name]
        else:
            tables = [
                s
                for s in db.table_names() + db.view_names()
                if not any(["_fts_" in s, s.endswith("_fts"), s.startswith("sqlite_")])
            ]
            if table_index is not None:
                tables = [tables[table_index]]

        for table in tables:
            df = pd.DataFrame(db[table].rows_where(offset=start_row, limit=end_row, order_by=order_by))
            dfs.append(NDF(table, df))
        db.close()
    elif mimetype in (
        "excel",
        "application/vnd.ms-excel",
        "excel spreadsheet subheader",
        "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
    ):
        excel_data = pd.read_excel(
            path,
            sheet_name=table_name or table_index,
            nrows=end_row,
            skiprows=start_row,
            header=None if skip_headers else 0,
        )

        if isinstance(excel_data, pd.DataFrame):
            worksheet_names = excel_data.index.levels[0]  # type: ignore
            for name in worksheet_names:
                df = excel_data.loc[name]
                dfs.append(NDF(name, df))
        else:
            for worksheet_name, df in excel_data.items():
                dfs.append(NDF(worksheet_name, df))
    elif mimetype in (
        "netcdf",
        "application/x-netcdf",
    ):
        import xarray as xr  # type: ignore

        ds = xr.open_dataset(path)
        df = ds.to_dataframe()
        dfs.append(NDF(None, df))
    elif mimetype in ("zarr",):
        import xarray as xr  # type: ignore

        ds = xr.open_zarr(path)
        df = ds.to_dataframe()
        dfs.append(NDF(None, df))
    elif mimetype in (
        "hdf",
        "application/x-hdf",
    ):
        df = pd.read_hdf(path, start=start_row, stop=end_row)
        dfs.append(NDF(None, df))
    elif mimetype in (
        "json",
        "application/json",
    ):
        df = pd.read_json(path, encoding=encoding)
        dfs.append(NDF(None, df))
    elif mimetype in ("jsonl", "json lines", "geojson lines"):
        df = pd.read_json(StringIO(path), nrows=end_row, lines=True, encoding=encoding)
        dfs.append(NDF(None, df))
    elif mimetype in (
        "csv",
        "text/csv",
    ):
        df = pd.read_csv(
            path, nrows=end_row, skiprows=start_row or 0, encoding=encoding, header=None if skip_headers else 0
        )
        dfs.append(NDF(None, df))
    elif mimetype in (
        "lines",
        "text/plain",
        "plaintext",
        "plain",
    ):
        df = pd.read_csv(
            path,
            names=["text"],
            sep="\r",  # just something to keep the parser busy
            nrows=end_row,
            skiprows=start_row or 0,
            encoding=encoding,
            header=None if skip_headers else 0,
        )
        dfs.append(NDF(None, df))
    elif mimetype in (
        "wsv",
        "text/wsv",
        "text/whitespace-separated-values",
    ):
        df = pd.read_csv(
            path,
            delim_whitespace=True,
            nrows=end_row,
            skiprows=start_row or 0,
            encoding=encoding,
            header=None if skip_headers else 0,
        )
        dfs.append(NDF(None, df))
    elif mimetype in (
        "tsv",
        "text/tsv",
        "text/tab-separated-values",
    ):
        df = pd.read_csv(
            path,
            delimiter="\t",
            nrows=end_row,
            skiprows=start_row or 0,
            encoding=encoding,
            header=None if skip_headers else 0,
        )
        dfs.append(NDF(None, df))
    elif mimetype in ("parq", "parquet", "application/parquet"):
        df = pd.read_parquet(path)
        dfs.append(NDF(None, df))
    elif mimetype in ("pkl", "pickle", "application/octet-stream"):
        df = pd.read_pickle(path)
        dfs.append(NDF(None, df))
    elif mimetype in (
        "html",
        "htm",
        "text/html",
        "html document",
    ):
        if path.startswith("http"):
            path = StringIO(web.extract_html(path))
        dfs.extend(NDF(None, df) for df in pd.read_html(path, skiprows=start_row, encoding=encoding))
    elif mimetype in ("stata",):
        df = pd.read_stata(path)
        dfs.append(NDF(None, df))
    elif mimetype in ("feather",):
        df = pd.read_feather(path)
        dfs.append(NDF(None, df))
    elif mimetype in ("orc",):
        df = pd.read_orc(path)
        dfs.append(NDF(None, df))
    elif "pdf" in mimetype:
        import camelot

        if " " in path:
            path = web.url_encode(path)  # camelot does not like spaces in URLs...

        for t in camelot.read_pdf(path, pages="all", suppress_stdout=False):  # type: ignore
            df = t.df
            if start_row:
                df = df.iloc[start_row:]
            if end_row:
                df = df.iloc[:end_row]
            if isinstance(df.columns, pd.RangeIndex):
                new_column_names = df.iloc[0]
                df = df.iloc[1:]
                df.columns = new_column_names
            df.columns = [" ".join(col.replace("\n", "").split()) for col in df.columns]
            dfs.append(df)
    elif "xml" in mimetype:
        df = pd.read_xml(path, encoding=encoding or "utf-8")
        dfs.append(NDF(None, df))
    else:
        msg = f"{path}: Unsupported file type: {mimetype}"
        raise ValueError(msg)

    if mimetype not in ("sqlite", "sqlite3", "sqlite database file") and table_index is not None:
        dfs = [dfs[table_index]]

    if skip_headers:
        for df_name, df in dfs:
            df.columns = [f"column{i}" for i in range(len(df.columns))]

    if join_tables:
        dfs = [NDF(dfs[0].df_name, pd.concat([t.df for t in dfs], axis=0, ignore_index=True))]

    dfs = [NDF(t.df_name if t.df_name else str(table_index_as_name), t.df) for table_index_as_name, t in enumerate(dfs)]

    if transpose:

        def transpose_df(df):
            df.loc[-1] = df.columns
            df = df.sort_index().reset_index(drop=True).T
            return df

        dfs = [NDF(t.df_name, transpose_df(t.df)) for t in dfs]

    return dfs


def get_filesize(d):
    try:
        stat = Path(d["path"]).stat()
        d["size"] = stat.st_size
        d["time_deleted"] = 0
    except FileNotFoundError:
        d["size"] = None
        if "time_deleted" not in d:
            d["time_deleted"] = consts.APPLICATION_START

    return d


def alt_name(file_path):
    file_path = Path(file_path)
    alternative_path = file_path
    counter = 1
    while alternative_path.exists():
        new_name = f"{file_path.stem}_{counter}{file_path.suffix}"
        alternative_path = file_path.with_name(new_name)
        counter += 1
    return str(alternative_path)



#!/usr/bin/python3
import re, urllib.parse

from xklb import usage
from xklb.utils import arggroups, argparse_utils, devices
from xklb.utils.log_utils import log


def parse_args(args):
    parser = argparse_utils.ArgumentParser(usage=usage.expand_links)
    parser.add_argument(
        "--search-urls", "-s", action="append", required=True, help='List of search URLs with a placeholder "%s".'
    )
    parser.add_argument("--browser", nargs="?", const="default")
    arggroups.debug(parser)

    arggroups.paths_or_stdin(parser)
    args = parser.parse_intermixed_args()
    arggroups.args_post(args, parser)

    return args


def expand_links(args=None):
    args = parse_args(args)

    links = []
    for query in args.paths:
        for search_url in args.search_urls:
            words = re.findall(r"\w+", query)
            log.debug("words: %s", words)
            words = [word for word in words if word]
            encoded_query = urllib.parse.quote(" ".join(words))

            if "%s" not in search_url:
                raise SystemExit(f"Error: No '%s' in provided search_url '{search_url}'")

            links.append(search_url.replace("%s", encoded_query))

    devices.browse(args.browser, links)



import argparse, json, os
from pathlib import Path

from xklb import usage
from xklb.mediadb import db_media
from xklb.utils import arggroups, argparse_utils


def parse_args() -> argparse.Namespace:
    parser = argparse_utils.ArgumentParser(usage=usage.nicotine_import)
    parser.add_argument("--track-deleted", default=True, action=argparse.BooleanOptionalAction)
    arggroups.debug(parser)

    arggroups.database(parser)
    arggroups.paths_or_stdin(parser)
    args = parser.parse_intermixed_args()
    arggroups.args_post(args, parser, create_db=True)

    return args


def from_nicotine_file_list_to_records(data):
    from yt_dlp.utils import traverse_obj

    result = []
    for entry in data:
        path = entry[0].replace("\\", os.sep)
        for song in entry[1]:
            song_info = {"path": f"{path}/{song[1]}", "size": song[2], "duration": traverse_obj(song, (4, "1"))}
            result.append(song_info)

    return result


def nicotine_import() -> None:
    args = parse_args()

    for path in args.paths:
        file_stats = Path(path).stat()
        time_created = int(file_stats.st_mtime) or int(file_stats.st_ctime)

        with open(path) as fp:
            data = json.load(fp)

        data = from_nicotine_file_list_to_records(data)
        data = [d | {"time_created": time_created, "time_deleted": 0} for d in data]

        db_media.update_media(args, data, mark_deleted=args.track_deleted and len(args.paths) == 1)



import re

environment_error = re.compile(
    "|".join(
        r""".*No space left on device
.*Transport endpoint is not connected
.*File name too long""".splitlines(),
    )
)


ignore_error = re.compile(
    "|".join(
        r""".*warning
.*deprecated""".splitlines(),
    )
)

unsupported_error = re.compile(
    "|".join(
        r""".*at least one.* received no packets
.*does not contain any stream
.*Unsupported codec""".splitlines(),
    ),
    flags=re.IGNORECASE,
)

file_error = re.compile(
    "|".join(
        r""".*detected.*low score
.*no decoder found for: none
.*Error initializing filters""".splitlines(),
    ),
    flags=re.IGNORECASE,
)



from pathlib import Path

from xklb import usage
from xklb.utils import arggroups, argparse_utils, file_utils, pd_utils, web
from xklb.utils.log_utils import log


def parse_args():
    parser = argparse_utils.ArgumentParser(usage=usage.tables_add)
    arggroups.table_like(parser)
    parser.set_defaults(end_row="inf")
    parser.add_argument("--sort", "-u", default="random()")

    parser.add_argument(
        "--primary-keys", "--pk", action=argparse_utils.ArgparseList, help="Comma separated primary keys"
    )
    parser.add_argument(
        "--business-keys", "--bk", action=argparse_utils.ArgparseList, help="Comma separated business keys"
    )

    parser.add_argument("--upsert", action="store_true")
    parser.add_argument("--ignore", "--only-new-rows", action="store_true")

    parser.add_argument("--only-target-columns", action="store_true")
    parser.add_argument("--skip-columns", action=argparse_utils.ArgparseList)
    arggroups.debug(parser)

    arggroups.database(parser)
    arggroups.paths_or_stdin(parser)
    args = parser.parse_intermixed_args()
    arggroups.args_post(args, parser, create_db=True)

    arggroups.table_like_post(args)

    return args


def table_add(args, path):
    dfs = file_utils.read_file_to_dataframes(
        path,
        table_name=args.table_name,
        table_index=args.table_index,
        start_row=args.start_row,
        end_row=args.end_row,
        order_by=args.sort,
        encoding=args.encoding,
        mimetype=args.mimetype,
        join_tables=args.join_tables,
        transpose=args.transpose,
        skip_headers=args.skip_headers,
    )
    for i, (df_name, df) in enumerate(dfs):
        if args.table_rename:
            table = args.table_rename.replace("%n", df_name).replace("%i", str(i))
        elif args.table_name == "stdin":
            table = "stdin"
        elif df_name.isnumeric():
            table = Path(path).stem
            if len(dfs) > 1:
                table += df_name
        else:
            table = df_name
        log.info("[%s]: %s", path, table)

        df = pd_utils.rename_duplicate_columns(df)
        df[pd_utils.available_name(df, "source_path")] = "stdin" if args.table_name == "stdin" else path

        skip_columns = args.skip_columns
        primary_keys = args.primary_keys
        if args.business_keys:
            if not primary_keys:
                primary_keys = list(o.name for o in args.db[table].columns if o.is_pk)

            skip_columns = [*(args.skip_columns or []), *primary_keys]

        selected_columns = df.columns.to_list()
        if args.only_target_columns:
            target_columns = args.db[table].columns_dict
            selected_columns = [s for s in selected_columns if s in target_columns]
        if skip_columns:
            selected_columns = [s for s in selected_columns if s not in skip_columns]

        log.info("[%s]: %s", table, selected_columns)
        kwargs = {}
        if args.business_keys or primary_keys:
            source_table_pks = [s for s in (args.business_keys or primary_keys) if s in selected_columns]
            if source_table_pks:
                log.info("[%s]: Using %s as primary key(s)", table, ", ".join(source_table_pks))
                kwargs["pk"] = source_table_pks

        data = df.to_dict(orient="records")
        data = ({k: v for k, v in d.items() if k in selected_columns} for d in data)
        with args.db.conn:
            args.db[table].insert_all(
                data,
                alter=True,
                ignore=args.ignore,
                replace=not args.ignore,
                upsert=args.upsert,
                **kwargs,
            )


def tables_add():
    args = parse_args()
    web.requests_session(args)  # configure session
    for path in args.paths:
        table_add(args, path)


if __name__ == "__main__":
    tables_add()



import csv, math, os, sys, textwrap, time
from collections.abc import Callable
from datetime import datetime, timezone

import humanize
from tabulate import tabulate

from xklb.utils import consts, web
from xklb.utils.strings import duration, file_size, relative_datetime


def print_overwrite(*text):
    if consts.PYTEST_RUNNING or not sys.stdout.isatty():
        pass
    elif os.name == "posix":
        print("\r" + text[0], *text[1:], end="\033[K", flush=False)
    elif consts.IS_WINDOWS:
        print("\r" + text[0], *text[1:], end="", flush=False)
    else:
        print(text)


def table(tbl, **kwargs) -> None:
    try:
        print(tabulate(tbl, tablefmt=consts.TABULATE_STYLE, headers="keys", showindex=False, **kwargs))
    except BrokenPipeError:
        sys.stdout = None
        sys.exit(141)


def pipe_print(*args) -> None:
    try:
        print(*args, flush=True)
    except BrokenPipeError:
        sys.stdout = None
        sys.exit(141)


def pipe_lines(x) -> None:
    try:
        sys.stdout.writelines(x)  # must include line endings
    except BrokenPipeError:
        sys.stdout = None
        sys.exit(141)


def eta(cur_iter, max_iter, start_time=None):
    if start_time is None:
        start_time = consts.APPLICATION_START

    elapsed = time.time() - start_time
    estimated = (elapsed / max(cur_iter, 1)) * (max_iter)

    finish_time = start_time + estimated
    return "ETA: " + relative_datetime(finish_time)


def write_csv_to_stdout(data):
    fieldnames = data[0].keys()
    writer = csv.DictWriter(sys.stdout, fieldnames=fieldnames)
    writer.writeheader()
    writer.writerows(data)


def path_fill(text, percent=None, width=None):
    if percent:
        width = max(10, int(percent * (consts.TERMINAL_SIZE.columns / 80)))
    lines = []
    current_line = ""
    for char in str(text):
        if char == "\r":
            continue  # Ignore carriage return character
        elif char == "\n":
            lines.append(current_line)
            current_line = ""
        else:
            current_line += char
            if len(current_line) == width:
                lines.append(current_line)
                current_line = ""
    if current_line:
        lines.append(current_line)
    return "\n".join(lines)


def transform_column(transform_func: Callable):
    def transform(tbl: list[dict], col: str, **kwargs) -> list[dict]:
        for d in tbl:
            if col in d and d[col] is not None:
                d[col] = transform_func(d[col], **kwargs)
        return tbl

    return transform


def transform_column_0null(transform_func: Callable):
    def transform(tbl: list[dict], col: str, **kwargs) -> list[dict]:
        for d in tbl:
            if col in d and d[col] is not None:
                if int(d[col]) == 0:
                    d[col] = None
                else:
                    d[col] = transform_func(d[col], **kwargs)
        return tbl

    return transform


def seconds_to_hhmmss(seconds):
    if seconds < 0:
        seconds = abs(seconds)
    seconds = int(seconds)

    hours = seconds // 3600
    minutes = (seconds % 3600) // 60
    seconds = seconds % 60

    formatted_time = f"{hours:>2}:{minutes:02d}:{seconds:02d}"
    if hours == 0:
        formatted_time = f"   {minutes:>2}:{seconds:02d}"

    return formatted_time


def unquote_url(val: str) -> str:
    if val.startswith("http"):
        val = web.safe_unquote(val)
    return val


col_duration = transform_column(duration)
col_hhmmss = transform_column(seconds_to_hhmmss)
col_unquote_url = transform_column(unquote_url)
col_resize = transform_column(path_fill)
col_resize_percent = transform_column(path_fill)

col_filesize = transform_column_0null(file_size)
col_naturaldate = transform_column_0null(
    lambda val: humanize.naturaldate(datetime.fromtimestamp(val, tz=timezone.utc).astimezone())
)
col_naturaltime = transform_column_0null(relative_datetime)


def wrap_paragraphs(text, width=80):
    paragraphs = text.split("\n\n")
    wrapped_paragraphs = [textwrap.fill(paragraph, width=width) for paragraph in paragraphs]
    formatted_text = "\n\n".join(wrapped_paragraphs)
    return formatted_text


def distribute_excess_width(max_col_widths, sep_char=4):
    existing_width = sum(max_col_widths.values()) + (len(max_col_widths) * sep_char)
    wide_cols = {
        k: width
        for k, width in max_col_widths.items()
        if width > 14 and k not in ("duration", "size", *consts.EPOCH_COLUMNS)
    }
    excess_width = 14 + existing_width - consts.TERMINAL_SIZE.columns

    if excess_width <= 0:
        return {}

    distributed_widths = {}
    for key, width in wide_cols.items():
        ratio = width / sum(wide_cols.values())
        subtract_width = math.ceil(ratio * excess_width)
        distributed_widths[key] = width - subtract_width

    return distributed_widths


def calculate_max_col_widths(data):
    max_col_widths = {}
    for row in data:
        for key, value in row.items():
            if isinstance(value, str):
                lines = value.splitlines()
                max_line_length = max(len(line) for line in lines or [""])
                max_col_widths[key] = max(max_col_widths.get(key) or 0, max_line_length, len(key))
            elif isinstance(value, list):
                max_value_length = max(len(str(item)) for item in value)
                max_col_widths[key] = max(max_col_widths.get(key) or 0, max_value_length)
            else:
                max_value_length = len(str(value))
                max_col_widths[key] = max(max_col_widths.get(key) or 0, max_value_length)

    return max_col_widths


def print_df(df):
    for col in df.select_dtypes(include=["category"]).columns:
        df[col] = df[col].astype("str")
    print()
    print(df.fillna("<NA>").to_markdown(tablefmt="github"))
    print()


def print_series(s):
    if len(s) > 0:
        print()
        print("\n".join([f"- {col}" for col in s]))
        print()



import os, re
from multiprocessing import TimeoutError as mp_TimeoutError
from pathlib import Path
from timeit import default_timer as timer

from xklb.createdb import av
from xklb.files import sample_hash
from xklb.folders import merge_mv
from xklb.mediafiles import process_ffmpeg, process_image
from xklb.utils import consts, file_utils, iterables, nums, objects, path_utils, processes, strings
from xklb.utils.consts import DBType
from xklb.utils.log_utils import log

REGEX_SENTENCE_ENDS = re.compile(r";|,|\.|\*|\n|\t")


def pop_substring_keys(e, key_substring):
    return [e.pop(k) for k in list(e.keys()) if key_substring in k]


def munge_book_tags(path) -> dict:
    try:
        import textract  # type: ignore
    except ModuleNotFoundError:
        print(
            "textract is required for text database creation: pip install textract; sudo dnf install libxml2-devel libxslt-devel antiword unrtf poppler-utils tesseract sox-plugins-nonfree sox libjpeg-devel swig",
        )
        raise
    try:
        tags = textract.process(path, language=os.getenv("TESSERACT_LANGUAGE"))
        tags = REGEX_SENTENCE_ENDS.split(tags.decode())
    except Exception as e:
        log.warning(e)
        log.error(f"Failed reading text. {path}")
        tags = []
    return {"tags": strings.combine(tags)}


munge_book_tags_slow = processes.with_timeout(350)(munge_book_tags)
munge_book_tags_fast = processes.with_timeout(70)(munge_book_tags)


def extract_metadata(mp_args, path) -> dict[str, str | int | None] | None:
    try:
        path.encode()
    except UnicodeEncodeError:
        log.error("Could not encode file path as UTF-8. Skipping %s", path)
        return None

    try:
        stat = os.stat(path, follow_symlinks=False)
    except FileNotFoundError:
        return None
    except OSError:
        log.exception("OSError: possible disk error; check dmesg. %s", path)
        return None
    except Exception as e:
        log.error(f"%s {path}", e)
        return None

    media = {
        "path": path,
        "size": stat.st_size,
        "type": file_utils.mimetype(path),
        "time_created": int(stat.st_ctime),
        "time_modified": int(stat.st_mtime) or consts.now(),
        "time_downloaded": consts.APPLICATION_START,
        "time_deleted": 0,
    }

    ext = path.rsplit(".", 1)[-1].lower()
    is_scan_all_files = getattr(mp_args, "scan_all_files", False)

    if media["type"] == "directory":
        return None

    if media["size"] == 0 or not Path(path).exists():
        return media

    if objects.is_profile(mp_args, DBType.audio) and (ext in consts.AUDIO_ONLY_EXTENSIONS or is_scan_all_files):
        media = av.munge_av_tags(mp_args, media)
    elif objects.is_profile(mp_args, DBType.video) and (ext in consts.VIDEO_EXTENSIONS or is_scan_all_files):
        media = av.munge_av_tags(mp_args, media)

    if not Path(path).exists():  # av.munge_av_tags might delete if unplayable or corruption exceeds threshold
        return media

    text_exts = consts.TEXTRACT_EXTENSIONS
    if mp_args.ocr:
        text_exts |= consts.OCR_EXTENSIONS
    if mp_args.speech_recognition:
        text_exts |= consts.SPEECH_RECOGNITION_EXTENSIONS
    if objects.is_profile(mp_args, DBType.text) and (ext in text_exts or is_scan_all_files):
        try:
            start = timer()
            if any([mp_args.ocr, mp_args.speech_recognition]):
                media |= munge_book_tags_slow(path)
            else:
                media |= munge_book_tags_fast(path)
        except mp_TimeoutError:
            log.warning(f"Timed out trying to read file. {path}")
        else:
            log.debug(f"{timer()-start} {path}")

    if getattr(mp_args, "hash", False) and media["type"] != "directory" and media["size"] > 0:
        media["hash"] = sample_hash.sample_hash_file(path)

    if getattr(mp_args, "copy", False) and not file_utils.is_file_open(path):
        dest_path = merge_mv.gen_rel_path(path, mp_args.copy, ":")
        if getattr(mp_args, "clean_path", True):
            dest_path = path_utils.clean_path(bytes(dest_path))
        else:
            dest_path = str(dest_path)
        file_utils.copy_file(path, dest_path, simulate=mp_args.simulate)
        path = media["path"] = dest_path

    if getattr(mp_args, "move", False) and not file_utils.is_file_open(path):
        dest_path = merge_mv.gen_rel_path(path, mp_args.move, ":")
        if getattr(mp_args, "clean_path", True):
            dest_path = path_utils.clean_path(bytes(dest_path))
        else:
            dest_path = str(dest_path)
        file_utils.rename_move_file(path, dest_path, simulate=mp_args.simulate)
        path = media["path"] = dest_path

    if getattr(mp_args, "process", False):
        if objects.is_profile(mp_args, DBType.audio) and Path(path).suffix not in [".opus", ".mka"]:
            result = process_ffmpeg.process_path(
                mp_args,
                path,
                split_longer_than=2160 if mp_args.split_longer_than is None and "audiobook" in path.lower() else None,
            )
            if result is None:
                return None
            path = media["path"] = str(result)
        elif objects.is_profile(mp_args, DBType.video) and Path(path).suffix not in [".av1.mkv"]:
            result = process_ffmpeg.process_path(mp_args, path)
            if result is None:
                return None
            path = media["path"] = str(result)
        elif objects.is_profile(mp_args, DBType.image) and Path(path).suffix not in [".avif", ".avifs"]:
            result = process_image.process_path(mp_args, path)
            if result is None:
                return None
            path = media["path"] = str(result)

    return media


def munge_image_tags(m: dict, e: dict) -> dict:
    chroma_subsample = nums.safe_int(
        iterables.safe_unpack(
            e.pop("File:YCbCrSubSampling", None),
            *pop_substring_keys(e, "YCbCrSubSampling"),
        )
    )
    if chroma_subsample == 0:
        chroma_subsample = None

    unit_x = iterables.safe_unpack(
        *pop_substring_keys(e, "XResolution"),
    )
    unit_y = iterables.safe_unpack(
        *pop_substring_keys(e, "YResolution"),
    )
    unit = iterables.safe_unpack(
        *pop_substring_keys(e, "ResolutionUnit"),
    )
    if unit == 0:
        unit = None
        unit_x = None if unit_x == 1 else unit_x
        unit_y = None if unit_y == 1 else unit_y

    m = {
        **m,
        "orientation": iterables.safe_unpack(
            *pop_substring_keys(e, "Orientation"),
        ),
        "width": iterables.safe_unpack(
            e.pop("File:ImageWidth", None),
            e.pop("Composite:ImageWidth", None),
            e.pop("EXIF:ImageWidth", None),
            e.pop("EXIF:ExifImageWidth", None),
            e.pop("PNG:ImageWidth", None),
            *pop_substring_keys(e, "ImageWidth"),
        ),
        "height": iterables.safe_unpack(
            e.pop("File:ImageHeight", None),
            e.pop("Composite:ImageHeight", None),
            e.pop("EXIF:ImageHeight", None),
            e.pop("EXIF:ExifImageHeight", None),
            e.pop("PNG:ImageHeight", None),
            *pop_substring_keys(e, "ImageHeight"),
        ),
        "chroma_subsample": chroma_subsample,
        "color_depth": iterables.safe_unpack(
            *pop_substring_keys(e, "ColorResolutionDepth"),
        ),
        "color_background": iterables.safe_unpack(
            *pop_substring_keys(e, "BackgroundColor"),
        ),
        "color_transparent": iterables.safe_unpack(
            *pop_substring_keys(e, "TransparentColor"),
        ),
        "longitude": iterables.safe_unpack(
            e.pop("Composite:GPSLongitude", None),
            *pop_substring_keys(e, "GPSLongitude"),
        ),
        "latitude": iterables.safe_unpack(
            e.pop("Composite:GPSLatitude", None),
            *pop_substring_keys(e, "GPSLatitude"),
        ),
        "focal_length": iterables.safe_unpack(
            e.pop("Composite:Lens", None),
            *pop_substring_keys(e, "Lens35efl"),
        ),
        "unit": unit,
        "unit_x": unit_x,
        "unit_y": unit_y,
        "exiftool_warning": strings.combine(*pop_substring_keys(e, "ExifTool:Warning")),
        "tags": strings.combine(
            *pop_substring_keys(e, "Headline"),
            *pop_substring_keys(e, "Title"),
            *pop_substring_keys(e, "ImageDescription"),
            *pop_substring_keys(e, "Caption"),
            *pop_substring_keys(e, "Artist"),
            *pop_substring_keys(e, "By-line"),
            *pop_substring_keys(e, "Credit"),
            *pop_substring_keys(e, "DocumentNotes"),
            *pop_substring_keys(e, "URL_List"),
            *pop_substring_keys(e, "Keywords"),
            *pop_substring_keys(e, "Make"),
            *pop_substring_keys(e, "Model"),
            *pop_substring_keys(e, "LensID"),
            *pop_substring_keys(e, "Creator"),
            *pop_substring_keys(e, "Software"),
        ),
    }

    return m


def extract_image_metadata_chunk(metadata: list[dict]) -> list[dict]:
    try:
        import exiftool
    except ModuleNotFoundError:
        print(
            "exiftool and PyExifTool are required for image database creation: sudo dnf install perl-Image-ExifTool && pip install PyExifTool",
        )
        raise

    chunk_paths = [d["path"] for d in metadata]
    try:
        with exiftool.ExifToolHelper() as et:
            exif = et.get_metadata(chunk_paths)
    except exiftool.exceptions.ExifToolExecuteError:
        log.exception("exifTool failed executing get_metadata %s", metadata)
        return metadata

    exif_enriched = []
    for m, e in zip(metadata, exif, strict=True):
        assert m["path"] == e.pop("SourceFile")

        try:
            m = munge_image_tags(m, e)
        except Exception as e:
            log.error("[%s]: %s", m["path"], e)
            # continue ?
        exif_enriched.append(m)

    return exif_enriched



import argparse, statistics
from collections import Counter
from functools import partial

import natsort
import regex as re
from natsort import ns

from xklb import usage
from xklb.tablefiles import mcda
from xklb.utils import arggroups, argparse_utils, consts, db_utils, iterables, printing, processes, strings
from xklb.utils.log_utils import log
from xklb.utils.objects import Reverser


def parse_args() -> argparse.Namespace:
    parser = argparse_utils.ArgumentParser(usage=usage.regex_sort)
    arggroups.text_filtering(parser)
    arggroups.regex_sort(parser)
    arggroups.debug(parser)

    parser.add_argument("input_path", nargs="?", type=argparse.FileType("r"), default="-")
    parser.add_argument("output_path", nargs="?")
    args = parser.parse_args()

    arggroups.regex_sort_post(args)
    arggroups.args_post(args, parser)
    return args


def line_splitter(regexs: list[re.Pattern], l: str) -> list[str]:
    words = [l]
    for rgx in regexs:
        new_words = []
        for word in words:
            new_words.extend(rgx.findall(word))
        words = new_words
    return words


def word_sorter(args, NS_OPTS, word_sorts: list[consts.WordSortOpt], corpus_stats: Counter, l: list[str]):
    if "lastindex" in word_sorts or "-lastindex" in word_sorts:
        rl = list(reversed(l))

    def gen_word_key(word_sorts, word):
        key_parts = []
        for s in word_sorts:
            reverse = False
            if s.startswith("-"):
                s = s.lstrip("-")
                reverse = True

            if s == "skip":
                val = None  # no sorting
            elif s == "len":
                val = len(word)
            elif s == "count":
                val = corpus_stats.get(word)
            elif s == "dup":
                val = (corpus_stats.get(word) or 0) > 1
            elif s == "unique":
                val = (corpus_stats.get(word) or 0) == 1
            elif s == "index":
                val = l.index(word)
            elif s == "lastindex":
                val = rl.index(word)
            elif s == "linecount":
                val = l.count(word)

            elif s in ("alpha", "python"):
                val = word
            elif s in ("natural", "natsort"):
                val = natsort.natsort_keygen(alg=NS_OPTS | ns.DEFAULT)(word)
            elif s in ("path", "nspath"):
                val = natsort.natsort_keygen(alg=NS_OPTS | ns.PATH)(word)
            elif s in ("locale", "human"):
                val = natsort.natsort_keygen(alg=NS_OPTS | ns.LOCALE)(word)
            elif s == "signed":
                val = natsort.natsort_keygen(alg=NS_OPTS | ns.REAL)(word)
            elif s == "os":
                val = natsort.os_sort_keygen()(word)
            else:
                raise NotImplementedError

            key_parts.append(Reverser(val) if reverse else val)

        log.debug("word_sorter xs: %s", key_parts)
        return tuple(key_parts)

    try:
        split_index = word_sorts.index("mcda")
    except ValueError:
        words = sorted(l, key=partial(gen_word_key, word_sorts))
    else:
        # Split the list into two parts
        before_mcda = word_sorts[:split_index]
        after_mcda = word_sorts[split_index + 1 :]  # remove "mcda" from word_sorts

        rank = mcda.mcda_sorted(args, keys=[gen_word_key(after_mcda, xs) for xs in l])
        log.debug("word_sorter mcda: %s", [l[i] for i in rank["original_index"]])
        words = sorted(l, key=lambda word: gen_word_key(before_mcda, word) + (rank["original_index"][l.index(word)],))

    log.debug(f"word_sorter: {words}")
    return words


def line_sorter(
    args,
    NS_OPTS,
    line_sorts: list[consts.LineSortOpt],
    corpus_stats: Counter,
    original_lines: list[str],
    corpus: list[list[str]],
):
    def gen_line_key(line_sorts, original_line, words):
        key_parts = []
        for s in line_sorts:
            reverse = False
            if s.startswith("-"):
                s = s.lstrip("-")
                reverse = True

            if s == "skip":
                val = None  # no sorting
            elif s == "line":
                val = original_line
            elif s == "count":
                val = len(words)
            elif s == "len":
                val = len("".join(words))
            elif s == "dup":
                val = sum((corpus_stats.get(word) or 0) > 1 for word in words)
            elif s == "unique":
                val = sum((corpus_stats.get(word) or 0) == 1 for word in words)
            elif s == "alldup":
                val = len(words) >= 1 and all((corpus_stats.get(word) or 0) > 1 for word in words)
            elif s == "allunique":
                val = len(words) >= 1 and all((corpus_stats.get(word) or 0) == 1 for word in words)
            elif s == "alldups":
                val = len(words) > 1 and all((corpus_stats.get(word) or 0) > 1 for word in words)
            elif s == "alluniques":
                val = len(words) > 1 and all((corpus_stats.get(word) or 0) == 1 for word in words)
            elif s == "sum":
                val = sum(corpus_stats.get(word) or 0 for word in words)
            elif s == "dupmax":
                val = max(corpus_stats.get(word) or 0 for word in words)
            elif s == "dupmin":
                val = min(corpus_stats.get(word) or 0 for word in words)
            elif s in ("dupavg", "dupmean"):
                val = statistics.mean(corpus_stats.get(word) or 0 for word in words)
            elif s == "dupmedian":
                try:
                    val = statistics.median(corpus_stats.get(word) or 0 for word in words)
                except statistics.StatisticsError:
                    val = None
            elif s == "dupmode":
                try:
                    val = statistics.mode(corpus_stats.get(word) or 0 for word in words)
                except statistics.StatisticsError:
                    val = None

            elif s in ("alpha", "python"):
                val = words
            elif s in ("natural", "natsort"):
                val = natsort.natsort_keygen(alg=NS_OPTS | ns.DEFAULT)(words)
            elif s in ("path", "nspath"):
                val = natsort.natsort_keygen(alg=NS_OPTS | ns.PATH)(words)
            elif s in ("locale", "human"):
                val = natsort.natsort_keygen(alg=NS_OPTS | ns.LOCALE)(words)
            elif s == "signed":
                val = natsort.natsort_keygen(alg=NS_OPTS | ns.REAL)(words)
            elif s == "os":
                val = natsort.os_sort_keygen()(words)
            else:
                raise NotImplementedError

            key_parts.append(Reverser(val) if reverse else val)
        return tuple(key_parts)

    try:
        split_index = line_sorts.index("mcda")
    except ValueError:
        before_mcda = line_sorts
        after_mcda = []
    else:
        before_mcda = line_sorts[:split_index]
        after_mcda = line_sorts[split_index + 1 :]  # remove "mcda" from line_sorts

    mcda_index = None
    if after_mcda:
        keys = [
            gen_line_key(after_mcda, original_line, words)
            for words, original_line in zip(corpus, original_lines, strict=True)
        ]
        rank = mcda.mcda_sorted(args, keys)
        mcda_index = rank["original_index"]

    line_sort_keys: list[tuple[tuple, str]] = []
    for line_idx, (words, original_line) in enumerate(zip(corpus, original_lines, strict=True)):
        line_sort_key = gen_line_key(before_mcda, original_line, words)
        if mcda_index is not None:
            line_sort_key += (mcda_index[line_idx],)
        log.debug(repr(line_sort_key) + "\t" + repr(words))
        line_sort_keys.append((line_sort_key, original_line))

    sorted_z = sorted(line_sort_keys, key=lambda y: y[0])
    lines = [x[1] for x in sorted_z]
    return lines


def prepare_corpus(corpus):
    words = list(iterables.flatten(corpus))
    if len(words) == 0:
        processes.exit_error("no words found. Check your regex! (and remove --duplicates --unique)")

    corpus_stats = Counter(words)
    return words, corpus_stats


def filter_corpus(corpus_stats, words, unique, dups):
    if len(words) == 0:
        return False
    elif unique is True and dups is None:  # --unique
        return any((corpus_stats.get(word) or 0) == 1 for word in words)
    elif unique is False and dups is None:  # --no-unique
        return not any((corpus_stats.get(word) or 0) == 1 for word in words)
    elif unique is True and dups is False:  # --unique --no-dups
        return all((corpus_stats.get(word) or 0) == 1 for word in words)
    elif unique is None and dups is True:  # --dups
        return any((corpus_stats.get(word) or 0) > 1 for word in words)
    elif unique is None and dups is False:  # --no-dups
        return not any((corpus_stats.get(word) or 0) > 1 for word in words)
    elif unique is False and dups is True:  # --no-unique --dups
        return all((corpus_stats.get(word) or 0) > 1 for word in words)
    elif unique is None and dups is None:  # no filtering
        return True
    elif unique is True and dups is True:  # no filtering
        return True
    elif unique is False and dups is False:  # no filtering
        return True
    raise NotImplementedError


def text_processor(args, lines):
    if args.stop_words is None:
        from xklb.data import wordbank

        stop_words = wordbank.stop_words
    else:
        stop_words = set(args.stop_words)

    NS_OPTS = ns.NUMAFTER | ns.NOEXP | ns.NANLAST
    if args.compat:
        NS_OPTS = NS_OPTS | ns.COMPATIBILITYNORMALIZE | ns.GROUPLETTERS

    corpus = []
    for l in lines:
        l = l.replace("http://", "", 1)
        l = l.replace("https://", "", 1)

        words = line_splitter(args.regexs, l)
        log.debug(f"line_splitter:    {words}")

        words = [s.lower() for s in words]
        words = [s for s in words if s not in stop_words]
        log.debug(f"stop_word_filter: {words}")

        corpus.append(words)
    words, corpus_stats = prepare_corpus(corpus)

    if args.unique is not None or args.duplicates is not None:
        filtered_indices = [
            i for i, words in enumerate(corpus) if filter_corpus(corpus_stats, words, args.unique, args.duplicates)
        ]
        lines = [lines[i] for i in filtered_indices]
        corpus = [corpus[i] for i in filtered_indices]
        words, corpus_stats = prepare_corpus(corpus)

    avg_word_len = sum(len(word) for word in words) / len(words) if words else 0
    max_word_len = max(len(word) for word in words) if words else 0
    min_word_len = min(len(word) for word in words) if words else 0

    log.info(f"Corpus stats: {min_word_len=} {avg_word_len=:.2f} {max_word_len=}")

    dup_words = set(word for word, count in corpus_stats.items() if count > 1)
    unique_words = set(word for word, count in corpus_stats.items() if count == 1)

    avg_dup_count_per_word = len(dup_words) / len(words)
    avg_dup_count_per_line = len(dup_words) / len(lines)
    avg_unique_count_per_word = len(unique_words) / len(words)
    avg_unique_count_per_line = len(unique_words) / len(lines)
    log.info(f"              {avg_dup_count_per_word=:.2f} {avg_dup_count_per_line=:.2f}")
    log.info(f"              {avg_unique_count_per_word=:.2f} {avg_unique_count_per_line=:.2f}")

    corpus = [word_sorter(args, NS_OPTS, args.word_sorts, corpus_stats, l) for l in corpus]

    # get original lines but sorted
    lines = line_sorter(args, NS_OPTS, args.line_sorts, corpus_stats, lines, corpus)
    return lines


def sort_dicts(args, media):
    search_columns = {
        col
        for _table, table_config in db_utils.config.items()
        if "search_columns" in table_config
        for col in table_config["search_columns"]
    }

    sentence_strings = list(
        strings.path_to_sentence(" ".join(str(v) for k, v in d.items() if v and k in search_columns)) for d in media
    )
    media_keyed = {line: d for line, d in zip(sentence_strings, media, strict=True)}

    lines = text_processor(args, sentence_strings)

    media = [media_keyed[p] for p in lines]
    return media


def regex_sort() -> None:
    args = parse_args()

    lines = args.input_path.readlines()
    args.input_path.close()

    lines = list(s.rstrip("\n") for s in lines if s.strip())
    lines = text_processor(args, lines)
    lines = (p + "\n" for p in lines)
    if args.output_path:
        with open(args.output_path, "w") as output_fd:
            output_fd.writelines(lines)
    else:
        printing.pipe_lines(lines)


if __name__ == "__main__":
    regex_sort()



import argparse, logging, os, sys
from timeit import default_timer

from IPython.core import ultratb
from IPython.terminal import debugger


def check_stdio():
    try:
        has_stdin = os.getpgrp() == os.tcgetpgrp(sys.stdin.fileno())
        has_stdout = os.getpgrp() == os.tcgetpgrp(sys.stdout.fileno())
    except Exception:
        has_stdin, has_stdout = False, False
    return has_stdin, has_stdout


def clamp_index(arr, idx):
    return arr[min(max(idx, 0), len(arr) - 1)]


def argparse_log() -> logging.Logger:
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument("--verbose", "-v", action="count", default=0)
    parser.add_argument("--no-pdb", action="store_true")
    args, _unknown = parser.parse_known_args()

    has_stdin, has_stdout = check_stdio()

    if args.verbose > 0 and has_stdin and has_stdout:
        sys.breakpointhook = debugger.set_trace
        if not args.no_pdb:
            sys.excepthook = ultratb.FormattedTB(
                mode="Verbose" if args.verbose > 1 else "Context",
                color_scheme="Neutral",
                call_pdb=True,
                debugger_cls=debugger.TerminalPdb,
            )

    log_levels = [logging.WARNING, logging.INFO, logging.DEBUG]

    if args.verbose > 3 or args.verbose == 0:
        logging.root.handlers = []  # clear any existing handlers
        logging.basicConfig(
            level=clamp_index(log_levels, args.verbose),
            format="%(message)s",
        )
    else:
        logging.basicConfig(format="%(message)s")

    logger = logging.getLogger("library")
    logger.setLevel(clamp_index(log_levels, args.verbose))
    return logger


log = argparse_log()


def gen_logging(name, original_generator):
    for x in original_generator:
        if name is not None:
            log.debug(f"{name}: {x}")
        else:
            log.debug(x)
        yield x


class Timer:
    def __init__(self):
        self.reset()

    def reset(self):
        self.start_time = default_timer()

    def elapsed(self):
        if not hasattr(self, "start_time"):
            raise RuntimeError("Timer has not been started.")
        end_time = default_timer()
        elapsed_time = end_time - self.start_time
        self.reset()
        return f"{elapsed_time:.4f}"



import sys

from xklb import usage
from xklb.playback import media_printer
from xklb.utils import arggroups, argparse_utils, file_utils, web


def parse_args():
    parser = argparse_utils.ArgumentParser(usage=usage.columns)
    arggroups.table_like(parser)
    parser.add_argument("--sort", "-u", default="random()")
    arggroups.debug(parser)

    arggroups.paths_or_stdin(parser)
    args = parser.parse_args()
    arggroups.args_post(args, parser)

    if "end_row" in args.defaults:
        args.end_row = str(args.start_row + 1) if args.start_row is not None else "2"
    arggroups.table_like_post(args)

    return args


def file_columns(args, path):
    import pandas as pd

    for df_name, df in file_utils.read_file_to_dataframes(
        path,
        table_name=args.table_name,
        table_index=args.table_index,
        start_row=args.start_row,
        end_row=args.end_row,
        order_by=args.sort,
        encoding=args.encoding,
        mimetype=args.mimetype,
        join_tables=args.join_tables,
        transpose=args.transpose,
        skip_headers=args.skip_headers,
    ):
        if (df.dtypes == "object").all():
            df = df.convert_dtypes()
        dc = [(col, df.dtypes[col]) for col in df.columns]
        df = pd.DataFrame(dc, columns=["name", "type"])

        if args.cols:
            df = df[args.cols]

        if args.to_json:
            df.to_json(sys.stdout, orient="records", lines=True)
        elif len(df.columns) == 1:
            series = df.iloc[:, 0]
            for value in series:
                print(value)
        elif args.print:
            media_printer.media_printer(args, df.to_dict(orient="records"))
        else:
            print(f"## {path}:{df_name}")
            print()
            print(df.to_markdown(tablefmt="github", index=False))
            print()


def columns():
    args = parse_args()
    web.requests_session(args)  # configure session
    for path in args.paths:
        file_columns(args, path)


if __name__ == "__main__":
    columns()



import json

from xklb import usage
from xklb.utils import arg_utils, arggroups, argparse_utils, objects, printing
from xklb.utils.log_utils import log


def parse_utils():
    parser = argparse_utils.ArgumentParser(usage=usage.combinations)
    arggroups.debug(parser)

    args, unknown_args = parser.parse_known_args()
    arggroups.args_post(args, parser)
    return args, unknown_args


def combinations():
    args, unknown_args = parse_utils()

    data = arg_utils.dict_of_lists_from_unknown_args(unknown_args)
    if not data:
        log.error("No data given via arguments")
        raise SystemExit(2)

    log.info(dict(data))
    for d in objects.product(**data):
        printing.pipe_lines(json.dumps(d) + "\n")



import re
from contextlib import suppress


def kebab_camel_snake(col):
    col = re.sub(r"(?<!^)(?=[A-Z])", "_", col)
    col = col.lower()
    col = re.sub(r"\s+", "_", col)
    col = re.sub(r"[\]\[\)\(\{\}]+", "", col)
    col = re.sub(r"[\:\-\.]+", "_", col)
    col = re.sub(r"__+", "_", col)
    return col


def columns_snake_case(df):
    df.columns = [kebab_camel_snake(x) if isinstance(x, str) else x for x in df.columns]
    return df


def convert_dtypes(df, clean=False):
    for col in df.columns:
        with suppress(Exception):
            if clean:
                df[col] = df[col].str.replace(r"\[.*|\(.*|\/.*", "", regex=True)
            df.loc[:, col] = df[col].str.replace(",", "").astype(float)

    df = df.convert_dtypes()
    return df


def rename_duplicate_columns(df):
    import pandas

    cols = pandas.Series(df.columns)
    for dup in cols[cols.duplicated()].unique():
        cols[cols[cols == dup].index.values.tolist()] = [
            dup + "_" + str(i) if i != 0 else dup for i in range(sum(cols == dup))
        ]
    df.columns = cols

    return df


def available_name(df, column_name):
    while column_name in df.columns:
        if "_" in column_name:
            base_name, suffix = column_name.rsplit("_", 1)
            column_name = f"{base_name}_{int(suffix) + 1}"
        else:
            column_name = f"{column_name}_1"
    return column_name



import sys

from xklb import usage
from xklb.playback import media_printer
from xklb.utils import arggroups, argparse_utils, file_utils, web
from xklb.utils.log_utils import check_stdio


def parse_args():
    parser = argparse_utils.ArgumentParser(usage=usage.markdown_tables)
    arggroups.table_like(parser)
    parser.add_argument("--sort", "-u", default="random()")
    parser.add_argument("--to-parquet", action="store_true", help="Write to Parquet")
    arggroups.debug(parser)

    arggroups.paths_or_stdin(parser)
    args = parser.parse_intermixed_args()
    arggroups.args_post(args, parser)

    arggroups.table_like_post(args)

    return args


def file_markdown(args, path):
    for df_name, df in file_utils.read_file_to_dataframes(
        path,
        table_name=args.table_name,
        table_index=args.table_index,
        start_row=args.start_row,
        end_row=args.end_row,
        order_by=args.sort,
        encoding=args.encoding,
        mimetype=args.mimetype,
        join_tables=args.join_tables,
        transpose=args.transpose,
        skip_headers=args.skip_headers,
    ):
        if args.cols:
            df = df[args.cols]

        if getattr(args, "repl", False):
            breakpoint()
        if args.to_json:
            df.to_json(sys.stdout, orient="records", lines=True)
        elif args.to_parquet:
            _has_stdin, has_stdout = check_stdio()
            if has_stdout:
                output_path = (args.table_name or df_name) + ".parquet"
                df.to_parquet(output_path, index=None, compression="zstd")
            else:
                sys.stdout.buffer.write(df.to_parquet(index=None, compression="zstd"))
        elif args.print:
            media_printer.media_printer(args, df.to_dict(orient="records"))
        else:
            if args.table_name == "stdin":
                print(f"## stdin:{df_name}")
            else:
                print(f"## {path}:{df_name}")
            print()
            print(df.to_markdown(tablefmt="github", index=False))
            print()


def markdown_tables():
    args = parse_args()
    web.requests_session(args)  # configure session
    for path in args.paths:
        file_markdown(args, path)


if __name__ == "__main__":
    markdown_tables()



import itertools, pickle
from functools import wraps
from html.parser import HTMLParser
from io import StringIO
from pathlib import Path

import regex

from xklb import usage
from xklb.data import wordbank
from xklb.utils import arggroups, argparse_utils, iterables, printing


def parse_args():
    parser = argparse_utils.ArgumentParser(usage=usage.nouns)
    parser.add_argument("--html-strip", "--strip-html", action="store_true", help="Strip HTML tags")
    parser.add_argument("--unique", "-u", action="store_true", help="Deduplicate output")
    parser.add_argument("--all", "-a", action="store_true", help="Show all output, even non-dictionary words")
    parser.add_argument(
        "--prepend",
        default=False,
        const="1",
        nargs="?",
        help="Characters to prepend to outputs. Numbers are A-Z combinatorial",
    )
    parser.add_argument(
        "--append",
        default=False,
        const="1",
        nargs="?",
        help="Characters to append to outputs. Numbers are A-Z combinatorial",
    )
    arggroups.debug(parser)

    arggroups.paths_or_stdin(parser)
    args = parser.parse_intermixed_args()
    arggroups.args_post(args, parser)

    return args


RE_NOUNS_SPLIT = regex.compile(
    r"(?= [a-z]|(?<!\b[A-Z][a-z]*) (?=[A-Z]))|[.?!,\/#$%\^&\*;:{}=\-_`~()]|\,|\'|\"|\^|‘|’|“|”|\n| -| :| _",
)


class MLStripper(HTMLParser):
    def __init__(self) -> None:
        super().__init__()
        self.reset()
        self.strict = False
        self.convert_charrefs = True
        self.text = StringIO()

    def handle_data(self, data) -> None:
        self.text.write(data)

    def get_data(self) -> str:
        return self.text.getvalue()


def strip_tags(html) -> str:
    s = MLStripper()
    s.feed(html)
    return s.get_data()


def is_num(s) -> bool:
    return s.replace(".", "", 1).replace("-", "", 1).isdigit()


def printer(part) -> None:
    if not part:
        return

    part_lookup = part.lower()
    if part_lookup in wordbank.stop_words or part_lookup in wordbank.prepositions or is_num(part):
        return

    printing.pipe_print(part)


def line_splitter(txt):
    parts = RE_NOUNS_SPLIT.split(txt)
    return [s.strip() for s in parts if s.strip()]


def generate_strings(length):
    if length == 1:
        for char in "abcdefghijklmnopqrstuvwxyz":
            yield char
    else:
        for prefix in generate_strings(length - 1):
            for char in "abcdefghijklmnopqrstuvwxyz":
                yield prefix + char


def nouns() -> None:
    args = parse_args()

    if not args.all:
        with (Path(__file__).parent / ".." / "data" / "dictionary.pkl").open("rb") as f:
            dictionary = pickle.load(f)

    def part_processor(args, parts):
        for part in parts:
            append_gen = []
            if args.append:
                if args.append.isnumeric():
                    append = int(args.append)
                    if args.all:
                        append_gen = generate_strings(append)
                    else:
                        append_gen = sorted(
                            s.split(part, 1)[1]
                            for s in dictionary
                            if s.startswith(part) and len(s) == len(part) + append
                        )
                else:
                    append_gen = [args.append]

            prepend_gen = []
            if args.prepend:
                if args.prepend.isnumeric():
                    prepend = int(args.prepend)
                    if args.all:
                        prepend_gen = generate_strings(prepend)
                    else:
                        prepend_gen = sorted(
                            s.rsplit(part, 1)[0]
                            for s in dictionary
                            if s.endswith(part) and len(s) == len(part) + prepend
                        )
                else:
                    prepend_gen = [args.prepend]

            if prepend_gen and append_gen:
                for pre, post in itertools.product(prepend_gen, append_gen):
                    yield pre + part + post
            elif prepend_gen:
                for pre in prepend_gen:
                    yield pre + part
            elif append_gen:
                for post in append_gen:
                    yield part + post
            else:
                yield part

    if args.unique:
        part_processor = iterables.return_unique(part_processor)
    if not args.all:
        part_processor_fn = part_processor

        @wraps(part_processor_fn)
        def check_dictionary(*args, **kwargs):
            for item in part_processor_fn(*args, **kwargs):
                if item in dictionary:
                    yield item

        part_processor = check_dictionary

    for line in args.paths:
        if args.html_strip:
            txt = strip_tags(txt)

        parts = line_splitter(line)
        parts = part_processor(args, parts)
        for part in parts:
            printer(part)



from pathlib import Path

from bs4 import BeautifulSoup, NavigableString

from xklb import usage
from xklb.createdb import fs_add_metadata
from xklb.utils import arg_utils, arggroups, argparse_utils, devices, iterables, printing, strings, web
from xklb.utils.log_utils import log


def parse_args():
    parser = argparse_utils.ArgumentParser(usage=usage.extract_text)
    arggroups.requests(parser)
    arggroups.selenium(parser)

    parser.add_argument("--local-html", action="store_true", help="Treat paths as Local HTML files")
    parser.add_argument("--skip-links", action="store_true")
    parser.add_argument("--download", "--save", "--write", action="store_true")

    arggroups.debug(parser)
    arggroups.paths_or_stdin(parser)
    args = parser.parse_intermixed_args()
    arggroups.args_post(args, parser)

    web.requests_session(args)  # prepare requests session
    arggroups.selenium_post(args)

    return args


def parse_text(args, html_content):
    soup = BeautifulSoup(html_content, "lxml")

    for item in soup.find_all():
        if item.name in ["meta"]:
            continue

        for descendant in item.descendants:
            if isinstance(descendant, NavigableString):
                parent = descendant.find_parent()
                parent_name = parent.name if parent else None

                if parent_name == "a":
                    if not args.skip_links:
                        yield strings.un_paragraph(descendant.get_text(strip=True))
                else:
                    yield strings.un_paragraph(descendant.get_text(strip=True))


def get_text(args, url):
    is_error = False
    if not args.local_html and not url.startswith("http") and Path(url).is_file():
        text = fs_add_metadata.munge_book_tags_fast(url)
        if text:
            tags = text.get("tags")
            if tags:
                yield tags.replace(";", "\n")
        return None

    if args.selenium:
        web.selenium_get_page(args, url)

        if args.manual:
            while devices.confirm("Extract HTML from browser?"):
                markup = web.selenium_extract_html(args.driver)
                yield from parse_text(args, markup)
        else:
            for markup in web.infinite_scroll(args.driver):
                yield from parse_text(args, markup)
    else:
        if args.local_html:
            with open(url) as f:
                markup = f.read()
            url = "file://" + url
        else:
            try:
                r = web.session.get(url, timeout=120)
            except Exception:
                log.exception("Could not get a valid response from the server")
                return None
            if r.status_code == 404:
                log.warning("404 Not Found Error: %s", url)
                is_error = True
            else:
                r.raise_for_status()
            markup = r.content

        yield from parse_text(args, markup)

    web.sleep(args)

    if is_error:
        return None


def extract_text() -> None:
    args = parse_args()

    if args.selenium:
        web.load_selenium(args)
    try:
        for url in arg_utils.gen_paths(args):
            output_lines = []
            for s in iterables.return_unique(get_text)(args, url):
                if s is None:
                    break

                if args.download:
                    output_lines.append(s)
                else:
                    printing.pipe_print(s)

            if args.download:
                save_path = web.url_to_local_path(url)
                Path(save_path).parent.mkdir(exist_ok=True, parents=True)
                with open(save_path, "w") as f:
                    f.writelines(s + "\n" for s in output_lines)
    finally:
        if args.selenium:
            web.quit_selenium(args)


if __name__ == "__main__":
    extract_text()



import argparse
from collections import defaultdict

from xklb import usage
from xklb.playback import media_printer
from xklb.utils import arggroups, argparse_utils, consts, db_utils, nums, sqlgroups


def parse_args() -> argparse.Namespace:
    parser = argparse_utils.ArgumentParser(usage=usage.download_status)
    arggroups.sql_fs(parser)
    parser.set_defaults(print="p")

    arggroups.download(parser)

    arggroups.debug(parser)
    arggroups.database(parser)

    parser.set_defaults(fts=False)
    args = parser.parse_args()
    arggroups.args_post(args, parser)

    arggroups.sql_fs_post(args)
    return args


def download_status() -> None:
    args = parse_args()

    query, bindings = sqlgroups.construct_download_query(args, dl_status=True)
    media = args.db.query(query, bindings)

    extractor_stats = defaultdict(
        lambda: {
            "never_attempted": 0,
            "failed_recently": 0,
            "retry_queued": 0,
            "retries_exceeded": 0,
            "downloaded_recently": 0,
        }
    )

    retry_delay = nums.human_to_seconds(args.retry_delay)

    for m in media:
        extractor_key = m.get("extractor_key", "Playlist-less media")

        if "download_attempts" in m and (m["download_attempts"] or 0) > args.download_retries:
            extractor_stats[extractor_key]["retries_exceeded"] += 1
        elif (m.get("time_downloaded") or 0) > 0 or (
            not m["path"].startswith("http") and (m.get("webpath") or "").startswith("http")
        ):
            if (m["time_downloaded"] + retry_delay) >= consts.APPLICATION_START:
                extractor_stats[extractor_key]["downloaded_recently"] += 1
        elif m["path"].startswith("http"):
            if "time_modified" in m:
                if (m["time_modified"] or 0) > 0 and (m["time_modified"] + retry_delay) < consts.APPLICATION_START:
                    extractor_stats[extractor_key]["retry_queued"] += 1
                elif (m["time_modified"] or 0) > 0 and (m["time_modified"] + retry_delay) >= consts.APPLICATION_START:
                    extractor_stats[extractor_key]["failed_recently"] += 1
                else:  # time_modified == 0
                    extractor_stats[extractor_key]["never_attempted"] += 1
            else:
                extractor_stats[extractor_key]["never_attempted"] += 1

    media = [{"extractor_key": extractor_key, **d} for extractor_key, d in extractor_stats.items()]
    media = sorted(media, key=lambda x: (-x["never_attempted"], -x["retry_queued"], x["extractor_key"]))

    media_printer.media_printer(args, media, units="extractors")

    if "error" in db_utils.columns(args, "media") and args.verbose >= consts.LOG_INFO:
        query = """
        select error, count(*) count
        from media
        where error is not null
        group by 1
        order by 2 DESC
        """
        errors = list(args.db.query(query))

        small_group = errors[:20][-1]["count"]
        common_errors = []
        other_errors = []
        for error in errors:
            if error["count"] < small_group:
                other_errors.append(error)
            else:
                common_errors.append(error)

        common_errors.append({"error": "Other", "count": len(other_errors)})
        media_printer.media_printer(args, common_errors)
        print(f"Total errors: {sum(d['count'] for d in errors)}")



import datetime

from dateutil.parser import parse

from xklb import usage
from xklb.utils import arggroups, argparse_utils, nums, strings
from xklb.utils.log_utils import log


def print_timestamp(n):
    nonzero_denominator = n % 1
    print(n if nonzero_denominator else int(n))


def timestamps(defaults_override=None, usage_override=None):
    parser = argparse_utils.ArgumentParser(usage=usage_override or usage.timestamps)
    parser.add_argument("--from-unix", "--unix", "-u", action="store_true", help="Parse from UNIX time")
    parser.add_argument("--from-timezone", "--from-tz", "-fz", help="Convert from timezone")

    parser.add_argument(
        "--month-day-year",
        "-m-d-y",
        action="store_true",
        help="""Parse ambiguous 3-integer date as MDY (default)
Example: 01/10/05
MDY  2005-01-10
DMY  2005-10-01
YMD  2001-10-05
YDM  2001-05-10

Example: July 8th, 2009
MDY  07/08/09
DMY  08/07/09
YMD  09/07/08
YDM  09/08/07
""",
    )
    parser.add_argument("--day-month-year", "-d-m-y", action="store_true", help="Parse ambiguous 3-integer date as DMY")
    parser.add_argument("--year-month-day", "-y-m-d", action="store_true", help="Parse ambiguous 3-integer date as YMD")
    parser.add_argument("--year-day-month", "-y-d-m", action="store_true", help="Parse ambiguous 3-integer date as YDM")

    parser.add_argument(
        "--to-date-only", "--date-only", "--date", "-d", action="store_true", help="Format the output as only dates"
    )
    parser.add_argument(
        "--to-time-only", "--time-only", "--time", "-t", action="store_true", help="Format the output as only times"
    )
    parser.add_argument("--to-unix", "-U", action="store_true", help="Format as UNIX time")
    parser.add_argument("--to-timezone", "--to-tz", "-tz", help="Convert to timezone")
    parser.add_argument("--print-timezone", "-TZ", action="store_true", help="Format with timezone")

    arggroups.debug(parser)

    parser.add_argument(
        "dates", nargs="*", default=argparse_utils.STDIN_DASH, action=argparse_utils.ArgparseArgsOrStdin
    )

    parser.set_defaults(**(defaults_override or {}))
    args = parser.parse_args()
    arggroups.args_post(args, parser)

    day_first = None
    year_first = None
    if args.month_day_year:
        day_first = False
        year_first = False
    elif args.day_month_year:
        day_first = True
        year_first = False
    elif args.year_month_day:
        day_first = False
        year_first = True
    elif args.year_day_month:
        day_first = True
        year_first = True

    if args.from_timezone:
        from_tzinfo = strings.timezone(args.from_timezone)
    elif args.from_unix:
        from_tzinfo = datetime.timezone.utc  # this is needed to set the initial timezone
    else:
        from_tzinfo = None  # localtime

    if args.to_timezone:
        to_tzinfo = strings.timezone(args.to_timezone)
    elif args.to_unix:
        to_tzinfo = datetime.timezone.utc  # not really necessary but you can see the symmetry
    else:
        to_tzinfo = None  # localtime

    for date_str in args.dates:
        if args.from_unix:
            date = datetime.datetime.fromtimestamp(nums.safe_float(date_str), tz=datetime.timezone.utc)  # type: ignore
            date = date.replace(tzinfo=from_tzinfo)
            log.debug("%s\t%s", date, date.tzinfo)
        else:
            date = parse(
                date_str,
                default=datetime.datetime.now().replace(day=1, hour=0, minute=0, second=0, microsecond=0),
                fuzzy=True,
                dayfirst=day_first,
                yearfirst=year_first,
            )
            log.debug("%s\t%s", date, date.tzinfo)
            if date.tzinfo is None:
                date = date.replace(tzinfo=from_tzinfo)  # naive datetime => timezone-aware datetime
                log.debug("%s\t%s", date, date.tzinfo)

        date = date.astimezone(tz=to_tzinfo)
        log.debug("%s\t%s", date, date.tzinfo)

        if args.to_unix:
            if args.to_time_only:  # datetime.time
                print_timestamp(date.hour * 3600 + date.minute * 60 + date.second + date.microsecond / 1e6)
            elif args.to_date_only:  # datetime.date
                print_timestamp(
                    datetime.datetime(
                        year=date.year,
                        month=date.month,
                        day=date.day,
                        hour=0,
                        minute=0,
                        microsecond=0,
                        tzinfo=date.tzinfo,
                    ).timestamp()
                )
            else:  # datetime.datetime
                print_timestamp(date.timestamp())
        else:
            if not args.print_timezone:
                date = date.replace(tzinfo=None)

            if args.to_time_only:
                date = date.time()
            elif args.to_date_only:
                date = date.date()

            print(date.isoformat())


def times():
    timestamps({"to_time_only": True}, usage_override=usage.times)


def dates():
    timestamps({"to_date_only": True}, usage_override=usage.dates)



import argparse, math, os, sqlite3

from xklb import usage
from xklb.mediadb import db_history
from xklb.mediafiles import process_ffmpeg, process_image
from xklb.utils import (
    arg_utils,
    arggroups,
    argparse_utils,
    consts,
    devices,
    file_utils,
    iterables,
    nums,
    printing,
    processes,
    sqlgroups,
    strings,
)
from xklb.utils.log_utils import log


def parse_args() -> argparse.Namespace:
    parser = argparse_utils.ArgumentParser(usage=usage.process_media)
    arggroups.sql_fs(parser)
    parser.set_defaults(
        local_media_only=True,
        hide_deleted=True,
        cols=["path", "type", "duration", "size", "video_count", "video_codecs", "audio_codecs"],
    )
    arggroups.history(parser)

    parser.add_argument(
        "--valid",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="Attempt to process files with valid metadata",
    )
    parser.add_argument(
        "--invalid",
        action=argparse.BooleanOptionalAction,
        default=False,
        help="Attempt to process files with invalid metadata",
    )

    parser.add_argument("--min-savings-video", type=nums.float_from_percent, default="3%")
    parser.add_argument("--min-savings-audio", type=nums.float_from_percent, default="10%")
    parser.add_argument("--min-savings-image", type=nums.float_from_percent, default="15%")

    parser.add_argument(
        "--source-audio-bitrate",
        type=nums.human_to_bits,
        default="256kbps",
        help="Used to estimate duration when files are invalid or inside of archives",
    )
    parser.add_argument(
        "--source-video-bitrate",
        type=nums.human_to_bits,
        default="1400kbps",
        help="Used to estimate duration when files are invalid or inside of archives",
    )

    parser.add_argument("--target-audio-bitrate", type=nums.human_to_bits, default="144kbps")
    parser.add_argument("--target-video-bitrate", type=nums.human_to_bits, default="800kbps")
    parser.add_argument("--target-image-size", type=nums.human_to_bytes, default="250KiB")
    parser.add_argument(
        "--transcoding-video-rate", type=float, default=1.8, help="Ratio of duration eg. 4x realtime speed"
    )
    parser.add_argument(
        "--transcoding-audio-rate", type=float, default=70, help="Ratio of duration eg. 100x realtime speed"
    )
    parser.add_argument("--transcoding-image-time", type=float, default=2.5, metavar="SECONDS")

    arggroups.process_ffmpeg(parser)
    arggroups.debug(parser)

    arggroups.database_or_paths(parser)
    args = parser.parse_intermixed_args()
    arggroups.args_post(args, parser)

    arggroups.sql_fs_post(args)
    arggroups.process_ffmpeg_post(args)

    return args


def collect_media(args) -> list[dict]:
    if args.database:
        db_history.create(args)

        try:
            media = list(args.db.query(*sqlgroups.media_sql(args)))
        except sqlite3.OperationalError:
            media = list(args.db.query(*sqlgroups.fs_sql(args, args.limit)))
    else:
        media = arg_utils.gen_d(args)
        media = [d if "size" in d else file_utils.get_filesize(d) for d in media]
    return media


def check_shrink(args, m) -> list:
    m["ext"] = os.path.splitext(m["path"])[1].lower().lstrip(".")
    filetype = (m.get("type") or "").lower()
    if (
        (filetype and (filetype.startswith("audio/") or " audio" in filetype))
        or m["ext"] in consts.AUDIO_ONLY_EXTENSIONS
    ) and (m.get("video_count") or 0) == 0:
        is_invalid = False
        m["media_type"] = "Audio"

        if m.get("compressed_size"):
            m["duration"] = m["size"] / args.source_audio_bitrate * 8

        if "duration" not in m:
            probe = processes.FFProbe(m["path"])
            m["duration"] = probe.duration
        if m["duration"] is None or not m["duration"] > 0:
            log.debug("[%s]: Invalid duration", m["path"])
            m["duration"] = m["size"] / args.source_audio_bitrate * 8
            is_invalid = True

        if (m.get("audio_codecs") or "") == "opus":
            log.debug("[%s]: Already opus", m["path"])
            return []

        future_size = int(m["duration"] * (args.target_audio_bitrate / 8))
        should_shrink_buffer = int(future_size * args.min_savings_audio)

        m["future_size"] = future_size
        m["savings"] = (m.get("compressed_size") or m["size"]) - future_size
        m["processing_time"] = math.ceil(m["duration"] / args.transcoding_audio_rate)

        can_shrink = m["size"] > (future_size + should_shrink_buffer)

        if is_invalid and args.invalid:
            return [m]
        elif args.valid and can_shrink:
            return [m]
        else:
            log.debug("[%s]: Skipping small file", m["path"])
    elif (
        (filetype and (filetype.startswith("image/") or " image" in filetype)) or m["ext"] in consts.IMAGE_EXTENSIONS
    ) and (m.get("duration") or 0) == 0:
        future_size = args.target_image_size
        should_shrink_buffer = int(future_size * args.min_savings_image)
        can_shrink = m["size"] > (future_size + should_shrink_buffer)

        m["media_type"] = "Image"
        m["future_size"] = future_size
        m["savings"] = (m.get("compressed_size") or m["size"]) - future_size
        m["processing_time"] = args.transcoding_image_time

        if can_shrink:
            return [m]
        log.debug("[%s]: Skipping small file", m["path"])
    elif (
        (filetype and (filetype.startswith("video/") or " video" in filetype)) or m["ext"] in consts.VIDEO_EXTENSIONS
    ) and (m.get("video_count") or 1) >= 1:
        is_invalid = False
        m["media_type"] = "Video"

        if m.get("compressed_size"):
            m["duration"] = m["size"] / args.source_video_bitrate * 8

        if "duration" not in m:
            probe = processes.FFProbe(m["path"])
            m["duration"] = probe.duration
        if m["duration"] is None or not m["duration"] > 0:
            log.debug("[%s]: Invalid duration", m["path"])
            m["duration"] = m["size"] / args.source_video_bitrate * 8
            is_invalid = True

        if (m.get("video_codecs") or "") == "av1":
            log.debug("[%s]: Already AV1", m["path"])
            return []

        future_size = int(m["duration"] * (args.target_video_bitrate / 8))
        should_shrink_buffer = int(future_size * args.min_savings_video)

        m["future_size"] = future_size
        m["savings"] = (m.get("compressed_size") or m["size"]) - future_size
        m["processing_time"] = math.ceil(m["duration"] / args.transcoding_video_rate)

        can_shrink = m["size"] > (future_size + should_shrink_buffer)

        if is_invalid and args.invalid:
            return [m]
        elif args.valid and can_shrink:
            return [m]
        else:
            log.debug("[%s]: Skipping small file", m["path"])
    elif (filetype and (filetype.startswith("archive/") or filetype.endswith("+zip") or " archive" in filetype)) or m[
        "ext"
    ] in consts.ARCHIVE_EXTENSIONS:
        contents = processes.lsar(m["path"])
        return [check_shrink(args, d) for d in contents]
    else:
        # TODO: pdf => avif
        # TODO: mobi, azw3, pdf => epub
        # TODO: csv, json => parquet
        log.warning("[%s]: Skipping unknown filetype %s", m["path"], filetype)
    return []


def process_media() -> None:
    args = parse_args()
    media = collect_media(args)

    media = iterables.conform(check_shrink(args, m) for m in media)
    media = sorted(
        media, key=lambda d: d["savings"] / (d["processing_time"] or args.transcoding_image_time), reverse=True
    )

    if not media:
        processes.no_media_found()

    summary = {}
    for m in media:
        media_key = f"{m['media_type']}: {m['ext']}"
        if m.get("compressed_size"):
            media_key += f" (archived)"

        if media_key not in summary:
            summary[media_key] = {
                "compressed_size": 0,
                "current_size": 0,
                "future_size": 0,
                "savings": 0,
                "processing_time": 0,
            }
        summary[media_key]["current_size"] += m["size"]
        summary[media_key]["future_size"] += m["future_size"]
        summary[media_key]["savings"] += m["savings"]
        summary[media_key]["compressed_size"] += m.get("compressed_size") or 0
        summary[media_key]["processing_time"] += m.get("processing_time") or 0

    summary = [{"media_key": k, **v} for k, v in summary.items()]
    savings = sum([m["savings"] for m in summary])
    if "processing_time" in summary[0]:
        processing_time = sum([m["processing_time"] for m in summary])

    summary = sorted(summary, key=lambda d: d["savings"], reverse=True)
    summary = iterables.list_dict_filter_bool(summary, keep_0=False)

    for t in ["processing_time"]:
        summary = printing.col_duration(summary, t)
    for t in ["current_size", "future_size", "compressed_size", "savings"]:
        summary = printing.col_filesize(summary, t)
    printing.table(summary)
    print()

    print("Estimated processing time:", strings.duration(processing_time))
    print("Estimated savings:", strings.file_size(savings))

    new_free_space = 0
    if devices.confirm(f"Proceed?"):
        for m in media:
            log.info(
                "%s freed. Processing %s (%s)",
                strings.file_size(new_free_space),
                m["path"],
                strings.file_size(m["size"]),
            )

            if m.get("compressed_size"):
                if os.path.exists(m["archive_path"]):
                    if args.simulate:
                        log.info("Unarchiving %s", m["archive_path"])
                    else:
                        processes.unar_delete(m["archive_path"])
                if not os.path.exists(m["path"]):
                    log.error("[%s]: FileNotFoundError from archive %s", m["path"], m["archive_path"])
                    continue
            else:
                if not os.path.exists(m["path"]):
                    log.error("[%s]: FileNotFoundError", m["path"])
                    m["time_deleted"] = consts.APPLICATION_START
                    continue

            if args.simulate:
                if m["media_type"] in ("Audio", "Video"):
                    log.info("FFMPEG processing %s", m["path"])
                elif m["media_type"] == "Image":
                    log.info("ImageMagick processing %s", m["path"])
                else:
                    raise NotImplementedError

                new_free_space += (m.get("compressed_size") or m["size"]) - m["future_size"]
            else:
                if m["media_type"] in ("Audio", "Video"):
                    new_path = process_ffmpeg.process_path(args, m["path"])
                elif m["media_type"] == "Image":
                    new_path = process_image.process_path(args, m["path"])
                else:
                    raise NotImplementedError

                if new_path is None:
                    m["time_deleted"] = consts.APPLICATION_START

                    new_free_space += m.get("compressed_size") or m["size"]
                else:
                    m["new_path"] = str(new_path)
                    m["new_size"] = os.stat(new_path).st_size

                    new_free_space += (m.get("compressed_size") or m["size"]) - m["new_size"]

                if args.database:
                    with args.db.conn:
                        if m.get("time_deleted"):
                            args.db.conn.execute(
                                "UPDATE media set time_deleted = ? where path = ?", [m["time_deleted"], m["path"]]
                            )
                        elif m.get("new_path") and m.get("new_path") != m["path"]:
                            args.db.conn.execute("DELETE FROM media where path = ?", [m["new_path"]])
                            args.db.conn.execute("UPDATE media set path = ? where path = ?", [m["new_path"], m["path"]])
                            args.db.conn.execute(
                                "UPDATE media SET path = ?, size = ? WHERE path = ?",
                                [m["new_path"], m["new_size"], m["path"]],
                            )



import sys

import requests

from xklb import usage
from xklb.createdb import gallery_backend, tube_backend
from xklb.mediadb import db_media
from xklb.mediafiles import process_ffmpeg, process_image
from xklb.playback import media_printer
from xklb.text import extract_links
from xklb.utils import (
    arg_utils,
    arggroups,
    argparse_utils,
    consts,
    db_utils,
    iterables,
    processes,
    sql_utils,
    strings,
    web,
)
from xklb.utils.consts import DBType
from xklb.utils.log_utils import log
from xklb.utils.sqlgroups import construct_download_query


def parse_args():
    parser = argparse_utils.ArgumentParser(usage=usage.download)
    arggroups.sql_fs(parser)

    arggroups.download(parser)
    arggroups.download_subtitle(parser)
    arggroups.requests(parser)
    arggroups.selenium(parser)
    arggroups.filter_links(parser)

    parser.add_argument("--same-domain", action="store_true", help="Choose a random domain to focus on")

    parser.add_argument("--live", action="store_true", help="Video: Allow live streams to be downloaded")

    parser.add_argument("--small", action="store_true", help="Video: Prefer 480p-like")

    parser.add_argument("--photos", action="store_true", help="Image: Only download JPG and WEBP")
    parser.add_argument("--drawings", action="store_true", help="Image: Only download PNG")
    parser.add_argument("--gifs", action="store_true", help="Image: Only download MP4 and GIFs")

    parser.add_argument("--links", action="store_true", help="Download media linked within pages")

    parser.add_argument("--process", action="store_true", help="Transcode images to AVIF and video/audio to AV1/Opus")
    arggroups.process_ffmpeg(parser)
    arggroups.debug(parser)

    arggroups.database(parser)
    arggroups.paths_or_stdin(parser, required=False)

    parser.set_defaults(fts=False)
    args, unk = parser.parse_known_intermixed_args()
    arggroups.args_post(args, parser, create_db=args.database and args.database.endswith((".db", ".sqlite")))

    if unk and not args.profile in (DBType.video, DBType.audio):
        parser.error(f"unrecognized arguments: {' '.join(unk)}")
    args.unk = unk

    if not args.profile and not args.print:
        log.error("Download profile must be specified. Use one of: --video OR --audio OR --image OR --filesystem")
        raise SystemExit(1)

    arggroups.sql_fs_post(args)
    arggroups.filter_links_post(args)
    web.requests_session(args)  # prepare requests session
    arggroups.selenium_post(args)
    arggroups.process_ffmpeg_post(args)
    return args


def download(args=None) -> None:
    if args:
        sys.argv = ["lb", *args]

    args = parse_args()

    m_columns = db_utils.columns(args, "media")

    if "limit" in args.defaults and "media" in args.db.table_names() and "webpath" in m_columns:
        if args.db.pop("SELECT 1 from media WHERE webpath is NULL and path in (SELECT webpath FROM media) LIMIT 1"):
            with args.db.conn:
                args.db.conn.execute(
                    """
                    DELETE from media WHERE path in (
                        SELECT webpath FROM media
                        WHERE error IS NULL OR error NOT LIKE 'Media check failed%'
                    ) AND webpath is NULL
                    """
                )

    args.blocklist_rules = []
    if "blocklist" in args.db.table_names():
        args.blocklist_rules = [{d["key"]: d["value"]} for d in args.db["blocklist"].rows]

    media = list(arg_utils.gen_d(args))
    if not media:
        query, bindings = construct_download_query(args)
        media = list(args.db.query(query, bindings))

    if not media:
        processes.no_media_found()

    if args.print:
        media_printer.media_printer(args, media)
        return

    get_inner_urls = iterables.return_unique(extract_links.get_inner_urls, lambda d: d["link"])
    for m in media:
        if args.blocklist_rules and sql_utils.is_blocked_dict_like_sql(m, args.blocklist_rules):
            continue

        if args.safe:
            if (args.profile in (DBType.audio, DBType.video) and not tube_backend.is_supported(m["path"])) or (
                args.profile in (DBType.image,) and not gallery_backend.is_supported(args, m["path"])
            ):
                log.info("[%s]: Skipping unsupported URL (safe_mode)", m["path"])
                continue

        # check if download already attempted recently by another process
        previous_time_attempted = m.get("time_modified") or consts.APPLICATION_START  # 0 is nullified
        if not args.force and "time_modified" in m_columns:
            d = args.db.pop_dict(
                f"""
                SELECT
                    time_modified
                    , time_deleted
                    {", download_attempts" if 'download_attempts' in m_columns else ', 0 as download_attempts'}
                FROM media
                WHERE path=?
                """,
                [m["path"]],
            )
            log.debug(d)
            if d:
                if d["time_deleted"]:
                    log.info(
                        "[%s]: Download was marked as deleted %s ago. Skipping!",
                        m["path"],
                        strings.duration(consts.now() - d["time_deleted"]),
                    )
                    continue
                elif d["time_modified"] > int(previous_time_attempted):
                    log.info(
                        "[%s]: Download already attempted %s ago. Skipping!",
                        m["path"],
                        strings.duration(consts.now() - d["time_modified"]),
                    )
                    continue
                elif d["download_attempts"] and d["download_attempts"] > args.download_retries:
                    log.info(
                        "[%s]: Download attempts exceed download retries limit. Skipping!",
                        m["path"],
                    )
                    continue

        try:  # attempt to download
            log.debug(m)

            if args.profile in (DBType.audio, DBType.video):
                tube_backend.download(args, m)
            elif args.profile == DBType.image:
                gallery_backend.download(args, m)
            elif args.profile == DBType.filesystem:
                original_path = m["path"]

                dl_paths = [original_path]
                if args.links:
                    dl_paths = []
                    try:
                        for link_dict in get_inner_urls(args, original_path):
                            dl_paths.append(link_dict["link"])
                    except requests.HTTPError as e:
                        log.warning(
                            "HTTPError %s. Recording download attempt: %s", e.response.status_code, original_path
                        )
                        db_media.download_add(
                            args,
                            webpath=original_path,
                            info=m,
                            error=str(e),
                            mark_deleted=e.response.status_code == 404,
                            delete_webpath_entry=False,
                        )
                        web.post_download(args)
                        continue

                if not dl_paths:
                    log.info("No relevant links in page. Recording download attempt: %s", original_path)
                    db_media.download_add(args, original_path, m, error="No relevant links in page")
                    web.post_download(args)
                    continue

                any_error = False
                for i, dl_path in enumerate(dl_paths):
                    error = None
                    try:
                        local_path = web.download_url(args, dl_path)
                    except RuntimeError as e:
                        local_path = None
                        error = str(e)

                    if local_path and args.process:
                        extension = local_path.rsplit(".", 1)[-1].lower()
                        if extension in consts.AUDIO_ONLY_EXTENSIONS | consts.VIDEO_EXTENSIONS:
                            result = process_ffmpeg.process_path(args, local_path)
                        elif extension in consts.IMAGE_EXTENSIONS:
                            result = process_image.process_path(args, local_path)

                        if result is not None:
                            local_path = str(result)

                    is_not_found = error is not None and "HTTPNotFound" in error
                    if error is not None and "HTTPNotFound" not in error:
                        any_error = True

                    db_media.download_add(
                        args,
                        webpath=original_path,
                        info=m,
                        local_path=local_path,
                        error=error,
                        mark_deleted=is_not_found,
                        delete_webpath_entry=(
                            not any_error if i == len(dl_paths) - 1 else False
                        ),  # only check after last download link was saved
                    )
            else:
                raise NotImplementedError

        except Exception:
            print("db:", args.database)
            raise



import shlex, subprocess
from pathlib import Path
from shutil import which

from xklb import usage
from xklb.createdb import fs_add_metadata
from xklb.utils import arggroups, argparse_utils, path_utils
from xklb.utils.arg_utils import gen_paths
from xklb.utils.log_utils import log


def parse_args():
    parser = argparse_utils.ArgumentParser(usage=usage.llm_map)
    parser.add_argument("--prompt", "-q", "--custom-prompt", help="Use a custom prompt")
    parser.add_argument(
        "--text", type=int, nargs="?", const=1500, help="Pass text file contents of each file to the LLM"
    )
    parser.add_argument("--rename", action="store_true", help="Use rename prompt")
    parser.add_argument("--output", help="The output CSV file to save the results.")
    arggroups.debug(parser)

    parser.add_argument(
        "--model",
        "-m",
        "--llamafile",
        help="The path to the llamafile to run. If llamafile is in your PATH then you can also specify a GGUF file.",
    )
    parser.add_argument(
        "--image-model",
        "--mmproj",
        help="The path to the LLaVA vision GGUF model. When specified, paths will be treated as image files",
    )
    parser.add_argument(
        "--llama-args", "--custom-args", type=shlex.split, default=[], help="Use custom llamafile arguments"
    )

    arggroups.paths_or_stdin(parser)
    args = parser.parse_intermixed_args()
    arggroups.args_post(args, parser)

    if args.prompt is None:
        if args.rename:
            args.prompt = "{name}Suggest an alternative, clean, filename based on the title and contents. Include the copyright year and any authors names in the format of SUBJECT.LASTNAME.YEAR.TITLE. Only say the filename and nothing else."
            args.llama_args += ["-n", "18"]
            if args.output is None:
                args.output = "llm_map_renames.csv"
        else:
            raise NotImplementedError

    args.exe = which("llamafile")
    if args.exe:
        args.llama_args += ["-m", args.model]
    else:
        args.exe = args.model

    if args.image_model:
        args.llama_args += ["--mmproj", args.image_model]

    if args.output is None:
        args.output = f"llm_map_{args.prompt}.csv"
    args.output = path_utils.clean_path(args.output.encode())
    return args


def run_llama_with_prompt(args, prompt):
    log.info(prompt)
    try:
        cmd = [args.exe, "--silent-prompt", *args.llama_args, "-p", prompt]
        result = subprocess.run(cmd, capture_output=True, text=True, check=True)
        return result.stdout.strip()
    except subprocess.CalledProcessError as e:
        print(f"Error running {cmd} for {prompt}: {e.stderr}")
        return None


def llm_map():
    args = parse_args()

    import pandas as pd

    results = []
    for path in gen_paths(args):
        prompt = args.prompt

        replacements = {
            "{path}": "Existing path: " + path,
            "{abspath}": "Existing path: " + str(Path(path).absolute()),
            "{name}": "Existing filename: " + Path(path).name,
            "{stem}": "Existing filename: " + Path(path).stem,
        }
        for k, v in replacements.items():
            prompt = prompt.replace(k, "\n" + v + "\n\n")

        if args.image_model:
            args.llama_args += ["--image", str(Path(path).absolute())]
        elif args.text:
            file_contents = fs_add_metadata.munge_book_tags_fast(path)
            if file_contents:
                file_contents = file_contents.get("tags")
            if file_contents:
                extract = file_contents[: args.text].replace(";", "\n")
                prompt = "File contents:\n" + extract + "\n\n" + prompt

        output = run_llama_with_prompt(args, prompt)
        if output and args.rename:
            output = output.replace("\\", "").replace("/", "")
            if output:
                output = path_utils.clean_path(bytes(Path(path).with_name(output)))

                ext = Path(path).suffix
                if not output.endswith(ext):
                    output += ext

        if output:
            results.append([path, output])

    df = pd.DataFrame(results, columns=["path", "output"])
    df.to_csv(args.output, index=False)
    print(f"Saved results to {args.output}")
